import pytest

from pypsrp.exceptions import (
    AuthenticationError,
    FragmentError,
    InvalidPipelineStateError,
    InvalidPSRPOperation,
    InvalidRunspacePoolStateError,
    SerializationError,
    WinRMError,
    WinRMTransportError,
    WSManFaultError,
)


def test_winrm_error():
    with pytest.raises(WinRMError) as exc:
        raise WinRMError("error msg")
    assert str(exc.value) == "error msg"


def test_authentication_error():
    with pytest.raises(AuthenticationError) as exc:
        raise AuthenticationError("auth error")
    assert str(exc.value) == "auth error"


def test_winrm_transport_error():
    with pytest.raises(WinRMTransportError) as exc:
        raise WinRMTransportError("proto", 1234, "response")
    assert str(exc.value) == "Bad PROTO response returned from the server. Code: 1234, Content: 'response'"
    assert exc.value.protocol == "proto"
    assert exc.value.code == 1234
    assert exc.value.response_text == "response"


def test_wsman_fault_error():
    with pytest.raises(WSManFaultError) as exc:
        raise WSManFaultError(1234, "machine", "reason", "provider", "path", "fault")
    assert (
        str(exc.value) == "Received a WSManFault message. (Code: 1234, "
        "Machine: machine, Reason: reason, "
        "Provider: provider, Provider Path: path, "
        "Provider Fault: fault)"
    )
    assert exc.value.code == 1234
    assert exc.value.machine == "machine"
    assert exc.value.reason == "reason"
    assert exc.value.provider == "provider"
    assert exc.value.provider_path == "path"
    assert exc.value.provider_fault == "fault"


def test_wsman_fault_error_empty():
    with pytest.raises(WSManFaultError) as exc:
        raise WSManFaultError(None, None, None, None, None, None)
    assert str(exc.value) == "Received a WSManFault message. (No details returned by the server)"


def test_invalid_runspace_pool_state_error():
    with pytest.raises(InvalidRunspacePoolStateError) as exc:
        raise InvalidRunspacePoolStateError(0, [1, 2], "do action")
    assert (
        str(exc.value) == "Cannot 'do action' on the current state "
        "'BeforeOpen', expecting state(s): "
        "'Opening, Opened'"
    )


def test_invalid_pipeline_state_error():
    with pytest.raises(InvalidPipelineStateError) as exc:
        raise InvalidPipelineStateError(0, 1, "do action")
    assert str(exc.value) == "Cannot 'do action' on the current state 'NotStarted', expecting state(s): 'Running'"


def test_invalid_psrp_operation():
    with pytest.raises(InvalidPSRPOperation) as exc:
        raise InvalidPSRPOperation("invalid psrp operation")
    assert str(exc.value) == "invalid psrp operation"


def test_fragment_error():
    with pytest.raises(FragmentError) as exc:
        raise FragmentError("fragment error")
    assert str(exc.value) == "fragment error"


def test_serialization_error():
    with pytest.raises(SerializationError) as exc:
        raise SerializationError("serialization error")
    assert str(exc.value) == "serialization error"



# Copyright: (c) 2018, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import division

import ipaddress
import logging
import re
import typing
import uuid
import warnings
import xml.etree.ElementTree as ET
from xml.dom.minidom import Element

import requests
from requests.packages.urllib3.util.retry import Retry

from pypsrp._utils import get_hostname, to_string, to_unicode
from pypsrp.encryption import WinRMEncryption
from pypsrp.exceptions import (
    AuthenticationError,
    WinRMError,
    WinRMTransportError,
    WSManFaultError,
)
from pypsrp.negotiate import HTTPNegotiateAuth

try:
    from requests_credssp import HttpCredSSPAuth
except ImportError as err:  # pragma: no cover
    _requests_credssp_import_error = "Cannot use CredSSP auth as requests-credssp is not installed: %s" % err

    class HttpCredSSPAuth(object):  # type: ignore[no-redef] # https://github.com/python/mypy/issues/1153
        def __init__(self, *args, **kwargs):
            raise ImportError(_requests_credssp_import_error)


log = logging.getLogger(__name__)

SUPPORTED_AUTHS = ["basic", "certificate", "credssp", "kerberos", "negotiate", "ntlm"]

AUTH_KWARGS: typing.Dict[str, typing.List[str]] = {
    "certificate": ["certificate_key_pem", "certificate_pem"],
    "credssp": ["credssp_auth_mechanism", "credssp_disable_tlsv1_2", "credssp_minimum_version"],
    "negotiate": ["negotiate_delegate", "negotiate_hostname_override", "negotiate_send_cbt", "negotiate_service"],
}

# [MS-WSMV] 2.2.1 Namespaces
# https://msdn.microsoft.com/en-us/library/ee878420.aspx
NAMESPACES: typing.Dict[str, str] = {
    "s": "http://www.w3.org/2003/05/soap-envelope",
    "xs": "http://www.w3.org/2001/XMLSchema",
    "xsi": "http://www.w3.org/2001/XMLSchema-instance",
    "wsa": "http://schemas.xmlsoap.org/ws/2004/08/addressing",
    "wsman": "http://schemas.dmtf.org/wbem/wsman/1/wsman.xsd",
    "wsmid": "http://schemas.dmtf.org/wbem/wsman/identify/1/wsmanidentity.xsd",
    "wsmanfault": "http://schemas.microsoft.com/wbem/wsman/1/wsmanfault",
    "cim": "http://schemas.dmtf.org/wbem/wscim/1/common",
    "wsmv": "http://schemas.microsoft.com/wbem/wsman/1/wsman.xsd",
    "cfg": "http://schemas.microsoft.com/wbem/wsman/1/config",
    "sub": "http://schemas.microsoft.com/wbem/wsman/1/subscription",
    "rsp": "http://schemas.microsoft.com/wbem/wsman/1/windows/shell",
    "m": "http://schemas.microsoft.com/wbem/wsman/1/machineid",
    "cert": "http://schemas.microsoft.com/wbem/wsman/1/config/service/certmapping",
    "plugin": "http://schemas.microsoft.com/wbem/wsman/1/config/PluginConfiguration",
    "wsen": "http://schemas.xmlsoap.org/ws/2004/09/enumeration",
    "wsdl": "http://schemas.xmlsoap.org/wsdl",
    "wst": "http://schemas.xmlsoap.org/ws/2004/09/transfer",
    "wsp": "http://schemas.xmlsoap.org/ws/2004/09/policy",
    "wse": "http://schemas.xmlsoap.org/ws/2004/08/eventing",
    "i": "http://schemas.microsoft.com/wbem/wsman/1/cim/interactive.xsd",
    "xml": "http://www.w3.org/XML/1998/namespace",
    "pwsh": "http://schemas.microsoft.com/powershell",
}


class WSManAction(object):
    # WS-Management URIs
    GET = "http://schemas.xmlsoap.org/ws/2004/09/transfer/Get"
    GET_RESPONSE = "http://schemas.xmlsoap.org/ws/2004/09/transfer/GetResponse"
    PUT = "http://schemas.xmlsoap.org/ws/2004/09/transfer/Put"
    PUT_RESPONSE = "http://schemas.xmlsoap.org/ws/2004/09/transfer/PutResponse"
    CREATE = "http://schemas.xmlsoap.org/ws/2004/09/transfer/Create"
    CREATE_RESPONSE = "http://schemas.xmlsoap.org/ws/2004/09/transfer/CreateResponse"
    DELETE = "http://schemas.xmlsoap.org/ws/2004/09/transfer/Delete"
    DELETE_RESPONSE = "http://schemas.xmlsoap.org/ws/2004/09/transfer/DeleteResponse"
    ENUMERATE = "http://schemas.xmlsoap.org/ws/2004/09/enumeration/Enumerate"
    ENUMERATE_RESPONSE = "http://schemas.xmlsoap.org/ws/2004/09/enumeration/EnumerateResponse"
    PULL = "http://schemas.xmlsoap.org/ws/2004/09/enumeration/Pull"
    PULL_RESPONSE = "http://schemas.xmlsoap.org/ws/2004/09/enumeration/PullResponse"

    # MS-WSMV URIs
    COMMAND = "http://schemas.microsoft.com/wbem/wsman/1/windows/shell/Command"
    COMMAND_RESPONSE = "http://schemas.microsoft.com/wbem/wsman/1/windows/shell/CommandResponse"
    CONNECT = "http://schemas.microsoft.com/wbem/wsman/1/windows/shell/Connect"
    CONNECT_RESPONSE = "http://schemas.microsoft.com/wbem/wsman/1/windows/shell/ConnectResponse"
    DISCONNECT = "http://schemas.microsoft.com/wbem/wsman/1/windows/shell/Disconnect"
    DISCONNECT_RESPONSE = "http://schemas.microsoft.com/wbem/wsman/1/windows/shell/DisconnectResponse"
    RECEIVE = "http://schemas.microsoft.com/wbem/wsman/1/windows/shell/Receive"
    RECEIVE_RESPONSE = "http://schemas.microsoft.com/wbem/wsman/1/windows/shell/ReceiveResponse"
    RECONNECT = "http://schemas.microsoft.com/wbem/wsman/1/windows/shell/Reconnect"
    RECONNECT_RESPONSE = "http://schemas.microsoft.com/wbem/wsman/1/windows/shell/ReconnectResponse"
    SEND = "http://schemas.microsoft.com/wbem/wsman/1/windows/shell/Send"
    SEND_RESPONSE = "http://schemas.microsoft.com/wbem/wsman/1/windows/shell/SendResponse"
    SIGNAL = "http://schemas.microsoft.com/wbem/wsman/1/windows/shell/Signal"
    SIGNAL_RESPONSE = "http://schemas.microsoft.com/wbem/wsman/1/windows/shell/SignalResponse"


class WSMan(object):
    def __init__(
        self,
        server: str,
        max_envelope_size: int = 153600,
        operation_timeout: int = 20,
        port: typing.Optional[int] = None,
        username: typing.Optional[str] = None,
        password: typing.Optional[str] = None,
        ssl: bool = True,
        path: str = "wsman",
        auth: str = "negotiate",
        cert_validation: bool = True,
        connection_timeout: int = 30,
        encryption: str = "auto",
        proxy: typing.Optional[str] = None,
        no_proxy: bool = False,
        locale: str = "en-US",
        data_locale: typing.Optional[str] = None,
        read_timeout: int = 30,
        reconnection_retries: int = 0,
        reconnection_backoff: float = 2.0,
        **kwargs: typing.Any,
    ) -> None:
        """
        Class that handles WSMan transport over HTTP. This exposes a method per
        action that takes in a resource and the header metadata required by
        that resource.

        This is required by the pypsrp.shell.WinRS and
        pypsrp.powershell.RunspacePool in order to connect to the remote host.
        It uses HTTP(S) to send data to the remote host.

        https://msdn.microsoft.com/en-us/library/cc251598.aspx

        :param server: The hostname or IP address of the host to connect to
        :param max_envelope_size: The maximum size of the envelope that can be
            sent to the server. Use update_max_envelope_size() to query the
            server for the true value
        :param max_envelope_size: The maximum size of a WSMan envelope that
            can be sent to the server
        :param operation_timeout: Indicates that the client expects a response
            or a fault within the specified time.
        :param port: The port to connect to, default is 5986 if ssl=True, else
            5985
        :param username: The username to connect with
        :param password: The password for the above username
        :param ssl: Whether to connect over http or https
        :param path: The WinRM path to connect to
        :param auth: The auth protocol to use; basic, certificate, negotiate,
            credssp. Can also specify ntlm or kerberos to limit the negotiate
            protocol
        :param cert_validation: Whether to validate the server's SSL cert
        :param connection_timeout: The timeout for connecting to the HTTP
            endpoint
        :param read_timeout: The timeout for receiving from the HTTP endpoint
        :param encryption: Controls the encryption setting, default is auto
            but can be set to always or never
        :param proxy: The proxy URL used to connect to the remote host
        :param no_proxy: Whether to ignore any environment proxy vars and
            connect directly to the host endpoint
        :param locale: The wsmv:Locale value to set on each WSMan request. This
            specifies the language in which the client wants response text to
            be translated. The value should be in the format described by
            RFC 3066, with the default being 'en-US'
        :param data_locale: The wsmv:DataLocale value to set on each WSMan
            request. This specifies the format in which numerical data is
            presented in the response text. The value should be in the format
            described by RFC 3066, with the default being the value of locale.
        :param int reconnection_retries: Number of retries on connection
            problems
        :param float reconnection_backoff: Number of seconds to backoff in
            between reconnection attempts (first sleeps X, then sleeps 2*X,
            4*X, 8*X, ...)
        :param kwargs: Dynamic kwargs based on the auth protocol set
            # auth='certificate'
            certificate_key_pem: The path to the cert key pem file
            certificate_pem: The path to the cert pem file

            # auth='credssp'
            credssp_auth_mechanism: The sub auth mechanism to use in CredSSP,
                default is 'auto' but can be 'ntlm' or 'kerberos'
            credssp_disable_tlsv1_2: Use TLSv1.0 instead of 1.2
            credssp_minimum_version: The minimum CredSSP server version to
                allow

            # auth in ['negotiate', 'ntlm', 'kerberos']
            negotiate_send_cbt: Whether to send the CBT token on HTTPS
                connections, default is True

            # the below are only relevant when kerberos (or nego used kerb)
            negotiate_delegate: Whether to delegate the Kerb token to extra
                servers (credential delegation), default is False
            negotiate_hostname_override: Override the hostname used when
                building the server SPN
            negotiate_service: Override the service used when building the
                server SPN, default='WSMAN'
        """
        log.debug(
            "Initialising WSMan class with maximum envelope size of %d "
            "and operation timeout of %s" % (max_envelope_size, operation_timeout)
        )
        self.session_id = str(uuid.uuid4())
        self.locale = locale
        self.data_locale = self.locale if data_locale is None else data_locale
        self.transport = _TransportHTTP(
            server,
            port,
            username,
            password,
            ssl,
            path,
            auth,
            cert_validation,
            connection_timeout,
            encryption,
            proxy,
            no_proxy,
            read_timeout,
            reconnection_retries,
            reconnection_backoff,
            **kwargs,
        )
        self.max_envelope_size = max_envelope_size
        self.operation_timeout = operation_timeout

        # register well known namespace prefixes so ElementTree doesn't
        # randomly generate them, saving packet space
        for key, value in NAMESPACES.items():
            ET.register_namespace(key, value)

        # This is the approx max size of a Base64 string that can be sent in a
        # SOAP message payload (PSRP fragment or send input data) to the
        # server. This value is dependent on the server's MaxEnvelopSizekb
        # value set on the WinRM service and the default is different depending
        # on the Windows version. Server 2008 (R2) detaults to 150KiB while
        # newer hosts are 500 KiB and this can be configured manually. Because
        # we don't know the OS version before we connect, we set the default to
        # 150KiB to ensure we are compatible with older hosts. This can be
        # manually adjusted with the max_envelope_size param which is the
        # MaxEnvelopeSizekb value * 1024. Otherwise the
        # update_max_envelope_size() function can be called and it will gather
        # this information for you.
        self.max_payload_size = self._calc_envelope_size(max_envelope_size)

    def __enter__(self) -> "WSMan":
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def command(
        self,
        resource_uri: str,
        resource: ET.Element,
        option_set: typing.Optional["OptionSet"] = None,
        selector_set: typing.Optional["SelectorSet"] = None,
        timeout: typing.Optional[int] = None,
    ) -> ET.Element:
        res = self.invoke(WSManAction.COMMAND, resource_uri, resource, option_set, selector_set, timeout)
        return res.find("s:Body", namespaces=NAMESPACES)  # type: ignore[return-value] # WSMan always has this present

    def connect(
        self,
        resource_uri: str,
        resource: ET.Element,
        option_set: typing.Optional["OptionSet"] = None,
        selector_set: typing.Optional["SelectorSet"] = None,
        timeout: typing.Optional[int] = None,
    ) -> ET.Element:
        res = self.invoke(WSManAction.CONNECT, resource_uri, resource, option_set, selector_set, timeout)
        return res.find("s:Body", namespaces=NAMESPACES)  # type: ignore[return-value] # WSMan always has this present

    def create(
        self,
        resource_uri: str,
        resource: ET.Element,
        option_set: typing.Optional["OptionSet"] = None,
        selector_set: typing.Optional["SelectorSet"] = None,
        timeout: typing.Optional[int] = None,
    ) -> ET.Element:
        res = self.invoke(WSManAction.CREATE, resource_uri, resource, option_set, selector_set, timeout)
        return res.find("s:Body", namespaces=NAMESPACES)  # type: ignore[return-value] # WSMan always has this present

    def disconnect(
        self,
        resource_uri: str,
        resource: ET.Element,
        option_set: typing.Optional["OptionSet"] = None,
        selector_set: typing.Optional["SelectorSet"] = None,
        timeout: typing.Optional[int] = None,
    ) -> ET.Element:
        res = self.invoke(WSManAction.DISCONNECT, resource_uri, resource, option_set, selector_set, timeout)
        return res.find("s:Body", namespaces=NAMESPACES)  # type: ignore[return-value] # WSMan always has this present

    def delete(
        self,
        resource_uri: str,
        resource: typing.Optional[ET.Element] = None,
        option_set: typing.Optional["OptionSet"] = None,
        selector_set: typing.Optional["SelectorSet"] = None,
        timeout: typing.Optional[int] = None,
    ) -> ET.Element:
        res = self.invoke(WSManAction.DELETE, resource_uri, resource, option_set, selector_set, timeout)
        return res.find("s:Body", namespaces=NAMESPACES)  # type: ignore[return-value] # WSMan always has this present

    def enumerate(
        self,
        resource_uri: str,
        resource: typing.Optional[ET.Element] = None,
        option_set: typing.Optional["OptionSet"] = None,
        selector_set: typing.Optional["SelectorSet"] = None,
        timeout: typing.Optional[int] = None,
    ) -> ET.Element:
        res = self.invoke(WSManAction.ENUMERATE, resource_uri, resource, option_set, selector_set, timeout)
        return res.find("s:Body", namespaces=NAMESPACES)  # type: ignore[return-value] # WSMan always has this present

    def get(
        self,
        resource_uri: str,
        resource: typing.Optional[ET.Element] = None,
        option_set: typing.Optional["OptionSet"] = None,
        selector_set: typing.Optional["SelectorSet"] = None,
        timeout: typing.Optional[int] = None,
    ) -> ET.Element:
        res = self.invoke(WSManAction.GET, resource_uri, resource, option_set, selector_set, timeout)
        return res.find("s:Body", namespaces=NAMESPACES)  # type: ignore[return-value] # WSMan always has this present

    def pull(
        self,
        resource_uri: str,
        resource: typing.Optional[ET.Element] = None,
        option_set: typing.Optional["OptionSet"] = None,
        selector_set: typing.Optional["SelectorSet"] = None,
        timeout: typing.Optional[int] = None,
    ) -> ET.Element:
        res = self.invoke(WSManAction.PULL, resource_uri, resource, option_set, selector_set, timeout)
        return res.find("s:Body", namespaces=NAMESPACES)  # type: ignore[return-value] # WSMan always has this present

    def put(
        self,
        resource_uri: str,
        resource: typing.Optional[ET.Element] = None,
        option_set: typing.Optional["OptionSet"] = None,
        selector_set: typing.Optional["SelectorSet"] = None,
        timeout: typing.Optional[int] = None,
    ) -> ET.Element:
        res = self.invoke(WSManAction.PUT, resource_uri, resource, option_set, selector_set, timeout)
        return res.find("s:Body", namespaces=NAMESPACES)  # type: ignore[return-value] # WSMan always has this present

    def receive(
        self,
        resource_uri: str,
        resource: ET.Element,
        option_set: typing.Optional["OptionSet"] = None,
        selector_set: typing.Optional["SelectorSet"] = None,
        timeout: typing.Optional[int] = None,
    ) -> ET.Element:
        res = self.invoke(WSManAction.RECEIVE, resource_uri, resource, option_set, selector_set, timeout)
        return res.find("s:Body", namespaces=NAMESPACES)  # type: ignore[return-value] # WSMan always has this present

    def reconnect(
        self,
        resource_uri: str,
        resource: typing.Optional[ET.Element] = None,
        option_set: typing.Optional["OptionSet"] = None,
        selector_set: typing.Optional["SelectorSet"] = None,
        timeout: typing.Optional[int] = None,
    ) -> ET.Element:
        res = self.invoke(WSManAction.RECONNECT, resource_uri, resource, option_set, selector_set, timeout)
        return res.find("s:Body", namespaces=NAMESPACES)  # type: ignore[return-value] # WSMan always has this present

    def send(
        self,
        resource_uri: str,
        resource: ET.Element,
        option_set: typing.Optional["OptionSet"] = None,
        selector_set: typing.Optional["SelectorSet"] = None,
        timeout: typing.Optional[int] = None,
    ) -> ET.Element:
        res = self.invoke(WSManAction.SEND, resource_uri, resource, option_set, selector_set, timeout)
        return res.find("s:Body", namespaces=NAMESPACES)  # type: ignore[return-value] # WSMan always has this present

    def signal(
        self,
        resource_uri: str,
        resource: ET.Element,
        option_set: typing.Optional["OptionSet"] = None,
        selector_set: typing.Optional["SelectorSet"] = None,
        timeout: typing.Optional[int] = None,
    ) -> ET.Element:
        res = self.invoke(WSManAction.SIGNAL, resource_uri, resource, option_set, selector_set, timeout)
        return res.find("s:Body", namespaces=NAMESPACES)  # type: ignore[return-value] # WSMan always has this present

    def get_server_config(
        self,
        uri: str = "config",
    ) -> ET.Element:
        resource_uri = "http://schemas.microsoft.com/wbem/wsman/1/%s" % uri
        log.debug("Getting server config with URI %s" % resource_uri)
        return self.get(resource_uri)

    def update_max_payload_size(
        self,
        max_payload_size: typing.Optional[int] = None,
    ) -> None:
        """
        Updates the MaxEnvelopeSize set on the current WSMan object for all
        future requests.

        :param max_payload_size: The max size specified in bytes, if not set
            then the max size if retrieved dynamically from the server
        """
        if max_payload_size is None:
            config = self.get_server_config()
            max_size_kb_et = config.find("cfg:Config/cfg:MaxEnvelopeSizekb", namespaces=NAMESPACES)
            max_size_kb = max_size_kb_et.text if max_size_kb_et is not None else ""
            max_payload_size = int(max_size_kb or "0") * 1024

        max_envelope_size = self._calc_envelope_size(max_payload_size)
        self.max_envelope_size = max_payload_size
        self.max_payload_size = max_envelope_size

    def invoke(
        self,
        action: str,
        resource_uri: str,
        resource: typing.Optional[ET.Element],
        option_set: typing.Optional["OptionSet"] = None,
        selector_set: typing.Optional["SelectorSet"] = None,
        timeout: typing.Optional[int] = None,
    ) -> ET.Element:
        """
        Send a generic WSMan request to the host.

        :param action: The action to run, this relates to the wsa:Action header
            field.
        :param resource_uri: The resource URI that the action relates to, this
          relates to the wsman:ResourceURI header field.
        :param resource: This is an optional xml.etree.ElementTree Element to
            be added to the s:Body section.
        :param option_set: a wsman.OptionSet to add to the request
        :param selector_set: a wsman.SelectorSet to add to the request
        :param timeout: Override the default wsman:OperationTimeout value for
            the request, this should be an int in seconds.
        :return: The ET Element of the response XML from the server
        """
        s = NAMESPACES["s"]
        envelope = ET.Element("{%s}Envelope" % s)

        message_id, header = self._create_header(action, resource_uri, option_set, selector_set, timeout)
        message_id = f"uuid:{message_id}"
        envelope.append(header)

        body = ET.SubElement(envelope, "{%s}Body" % s)
        if resource is not None:
            body.append(resource)

        xml = ET.tostring(envelope, encoding="utf-8", method="xml")

        try:
            response = self.transport.send(xml)
        except WinRMTransportError as err:
            try:
                # try and parse the XML and get the WSManFault
                raise self._parse_wsman_fault(err.response_text)
            except ET.ParseError:
                # no XML message is present so not a WSManFault error
                log.error(
                    "Failed to parse WSManFault message on WinRM error"
                    " response, raising original WinRMTransportError"
                )
                raise err

        response_xml = ET.fromstring(response)
        relates_to_et = response_xml.find("s:Header/wsa:RelatesTo", namespaces=NAMESPACES)
        relates_to = relates_to_et.text if relates_to_et is not None else ""

        if message_id != relates_to:
            raise WinRMError(
                "Received related id does not match related "
                "expected message id: Sent: %s, Received: %s" % (message_id, relates_to)
            )
        return response_xml

    def _calc_envelope_size(
        self,
        max_envelope_size: int,
    ) -> int:
        # get a mock Header which should cover most cases where large fragments
        # are used
        empty_uuid = "00000000-0000-0000-0000-000000000000"

        selector_set = SelectorSet()
        selector_set.add_option("ShellId", empty_uuid)
        _, header = self._create_header(
            WSManAction.SEND, "http://schemas.microsoft.com/wbem/wsman/1/windows/shell/cmd", selector_set=selector_set
        )

        # get a skeleton Body to calculate the size without the payload
        rsp = NAMESPACES["rsp"]
        send = ET.Element("{%s}Send" % rsp)
        ET.SubElement(send, "{%s}Stream" % rsp, Name="stdin", CommandId=empty_uuid).text = ""

        envelope = ET.Element("{%s}Envelope" % NAMESPACES["s"])
        envelope.append(header)
        envelope.append(send)
        envelope = ET.tostring(envelope, encoding="utf-8", method="xml")

        # add the Header and Envelope and pad some extra bytes to cover
        # slightly different scenarios, multiple options, different body types
        # while this isn't perfect it's better than wasting CPU cycles
        # calculating it per message and a few bytes don't make too much of a
        # difference
        envelope_size = len(envelope) + 256
        max_bytes_size = max_envelope_size - envelope_size

        # Data is sent as Base64 encoded which inflates the size, we need to
        # calculate how large that can be
        base64_size = int(max_bytes_size / 4 * 3)
        return base64_size

    def _create_header(
        self,
        action: str,
        resource_uri: str,
        option_set: typing.Optional["OptionSet"] = None,
        selector_set: typing.Optional["SelectorSet"] = None,
        timeout: typing.Optional[int] = None,
    ) -> typing.Tuple[str, ET.Element]:
        log.debug(
            "Creating WSMan header (Action: %s, Resource URI: %s, "
            "Option Set: %s, Selector Set: %s" % (action, resource_uri, option_set, selector_set)
        )
        s = NAMESPACES["s"]
        wsa = NAMESPACES["wsa"]
        wsman = NAMESPACES["wsman"]
        wsmv = NAMESPACES["wsmv"]
        xml = NAMESPACES["xml"]

        header = ET.Element("{%s}Header" % s)

        ET.SubElement(header, "{%s}Action" % wsa, attrib={"{%s}mustUnderstand" % s: "true"}).text = action

        ET.SubElement(
            header,
            "{%s}DataLocale" % wsmv,
            attrib={"{%s}mustUnderstand" % s: "false", "{%s}lang" % xml: self.data_locale},
        )

        ET.SubElement(
            header, "{%s}Locale" % wsman, attrib={"{%s}mustUnderstand" % s: "false", "{%s}lang" % xml: self.locale}
        )

        ET.SubElement(header, "{%s}MaxEnvelopeSize" % wsman, attrib={"{%s}mustUnderstand" % s: "true"}).text = str(
            self.max_envelope_size
        )

        message_id = str(uuid.uuid4()).upper()
        ET.SubElement(header, "{%s}MessageID" % wsa).text = "uuid:%s" % message_id

        ET.SubElement(header, "{%s}OperationTimeout" % wsman).text = "PT%sS" % str(timeout or self.operation_timeout)

        reply_to = ET.SubElement(header, "{%s}ReplyTo" % wsa)
        ET.SubElement(
            reply_to, "{%s}Address" % wsa, attrib={"{%s}mustUnderstand" % s: "true"}
        ).text = "http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous"

        ET.SubElement(header, "{%s}ResourceURI" % wsman, attrib={"{%s}mustUnderstand" % s: "true"}).text = resource_uri

        ET.SubElement(header, "{%s}SessionId" % wsmv, attrib={"{%s}mustUnderstand" % s: "false"}).text = (
            "uuid:%s" % str(self.session_id).upper()
        )

        ET.SubElement(header, "{%s}To" % wsa).text = self.transport.endpoint

        if option_set is not None:
            header.append(option_set.pack())

        if selector_set is not None:
            header.append(selector_set.pack())

        return message_id, header

    def close(self) -> None:
        self.transport.close()

    @staticmethod
    def _parse_wsman_fault(xml_text: str) -> WSManFaultError:
        xml = ET.fromstring(xml_text)
        code: typing.Optional[typing.Union[str, int]] = None
        reason = None
        machine = None
        provider = None
        provider_path = None
        provider_fault = None

        fault = xml.find("s:Body/s:Fault", namespaces=NAMESPACES)
        if fault is not None:
            code_info = fault.find("s:Code/s:Subcode/s:Value", namespaces=NAMESPACES)
            if code_info is not None:
                code = code_info.text
            else:
                code_info = fault.find("s:Code/s:Value", namespaces=NAMESPACES)
                if code_info is not None:
                    code = code_info.text

            reason_info = fault.find("s:Reason/s:Text", namespaces=NAMESPACES)
            if reason_info is not None:
                reason = reason_info.text

        wsman_fault = fault.find("s:Detail/wsmanfault:WSManFault", namespaces=NAMESPACES) if fault else None
        if wsman_fault is not None:
            code = wsman_fault.attrib.get("Code", code)
            machine = wsman_fault.attrib.get("Machine")

            message_info = wsman_fault.find("wsmanfault:Message", namespaces=NAMESPACES)
            if message_info is not None:
                # message may still not be set, fall back to the existing
                # reason value from the base soap Fault element
                reason = message_info.text or reason

            provider_info = wsman_fault.find("wsmanfault:Message/wsmanfault:ProviderFault", namespaces=NAMESPACES)
            if provider_info is not None:
                provider = provider_info.attrib.get("provider")
                provider_path = provider_info.attrib.get("path")
                provider_fault = provider_info.text

        # lastly try and cleanup the value of the parameters
        try:
            code = int(code or "")
        except (TypeError, ValueError):
            pass

        reason = reason.strip() if reason else None
        provider_fault = provider_fault.strip() if provider_fault else None

        return WSManFaultError(code, machine, reason, provider, provider_path, provider_fault)


class _WSManSet(object):
    def __init__(
        self,
        element_name: str,
        child_element_name: str,
        must_understand: bool,
    ) -> None:
        self.element_name = element_name
        self.child_element_name = child_element_name
        self.must_understand = must_understand
        self.values: typing.List[typing.Tuple[str, str, typing.Dict]] = []

    def __str__(self) -> str:
        # can't just str({}) as the ordering is important
        entry_values = []
        for value in self.values:
            entry_values.append("'%s': '%s'" % (value[0], value[1]))

        string_value = "{%s}" % ", ".join(entry_values)
        return string_value

    def add_option(
        self,
        name: str,
        value: str,
        attributes: typing.Optional[typing.Dict] = None,
    ) -> None:
        attributes = attributes if attributes is not None else {}
        self.values.append((name, value, attributes))

    def pack(self) -> ET.Element:
        s = NAMESPACES["s"]
        wsman = NAMESPACES["wsman"]
        element = ET.Element("{%s}%s" % (wsman, self.element_name))
        if self.must_understand:
            element.attrib["{%s}mustUnderstand" % s] = "true"

        for key, value, attributes in self.values:
            ET.SubElement(element, "{%s}%s" % (wsman, self.child_element_name), Name=key, attrib=attributes).text = str(
                value
            )

        return element


class OptionSet(_WSManSet):
    def __init__(self) -> None:
        super(OptionSet, self).__init__("OptionSet", "Option", True)


class SelectorSet(_WSManSet):
    def __init__(self) -> None:
        super(SelectorSet, self).__init__("SelectorSet", "Selector", False)


# this should not be used outside of this class
class _TransportHTTP(object):
    def __init__(
        self,
        server: str,
        port: typing.Optional[int] = None,
        username: typing.Optional[str] = None,
        password: typing.Optional[str] = None,
        ssl: bool = True,
        path: str = "wsman",
        auth: str = "negotiate",
        cert_validation: bool = True,
        connection_timeout: int = 30,
        encryption: str = "auto",
        proxy: typing.Optional[str] = None,
        no_proxy: bool = False,
        read_timeout: int = 30,
        reconnection_retries: int = 0,
        reconnection_backoff: float = 2.0,
        **kwargs: typing.Any,
    ) -> None:
        self.server = server
        self.port = port if port is not None else (5986 if ssl else 5985)
        self.username = username
        self.password = password
        self.ssl = ssl
        self.path = path

        if auth not in SUPPORTED_AUTHS:
            raise ValueError(
                "The specified auth '%s' is not supported, "
                "please select one of '%s'" % (auth, ", ".join(SUPPORTED_AUTHS))
            )
        self.auth = auth
        self.cert_validation = cert_validation
        self.connection_timeout = connection_timeout
        self.read_timeout = read_timeout
        self.reconnection_retries = reconnection_retries
        self.reconnection_backoff = reconnection_backoff

        # determine the message encryption logic
        if encryption not in ["auto", "always", "never"]:
            raise ValueError("The encryption value '%s' must be auto, always, or never" % encryption)
        enc_providers = ["credssp", "kerberos", "negotiate", "ntlm"]
        if ssl:
            # msg's are automatically encrypted with TLS, we only want message
            # encryption if always was specified
            self.wrap_required = encryption == "always"
            if self.wrap_required and self.auth not in enc_providers:
                raise ValueError(
                    "Cannot use message encryption with auth '%s', either set "
                    "encryption='auto' or use one of the following auth "
                    "providers: %s" % (self.auth, ", ".join(enc_providers))
                )
        else:
            # msg's should always be encrypted when not using SSL, unless the
            # user specifies to never encrypt
            self.wrap_required = not encryption == "never"
            if self.wrap_required and self.auth not in enc_providers:
                raise ValueError(
                    "Cannot use message encryption with auth '%s', either set "
                    "encryption='never', use ssl=True or use one of the "
                    "following auth providers: %s" % (self.auth, ", ".join(enc_providers))
                )
        self.encryption: typing.Optional[WinRMEncryption] = None

        self.proxy = proxy
        self.no_proxy = no_proxy

        self.certificate_key_pem: typing.Optional[str] = None
        self.certificate_pem: typing.Optional[str] = None
        for kwarg_list in AUTH_KWARGS.values():
            for kwarg in kwarg_list:
                setattr(self, kwarg, kwargs.get(kwarg, None))

        self.endpoint = self._create_endpoint(self.ssl, self.server, self.port, self.path)
        log.debug(
            "Initialising HTTP transport for endpoint: %s, user: %s, "
            "auth: %s" % (self.endpoint, self.username, self.auth)
        )
        self.session: typing.Optional[requests.Session] = None

        # used when building tests, keep commented out
        # self._test_messages = []

    def close(self) -> None:
        if self.session:
            self.session.close()

    def send(self, message: bytes) -> bytes:
        hostname = get_hostname(self.endpoint)
        if self.session is None:
            self.session = self._build_session()

            # need to send an initial blank message to setup the security
            # context required for encryption
            if self.wrap_required:
                request = requests.Request("POST", self.endpoint, data=None)
                prep_request = self.session.prepare_request(request)
                self._send_request(prep_request)

                protocol = WinRMEncryption.SPNEGO
                if isinstance(self.session.auth, HttpCredSSPAuth):
                    protocol = WinRMEncryption.CREDSSP
                elif self.session.auth.contexts[hostname].response_auth_header == "kerberos":  # type: ignore[union-attr] # This should not happen
                    # When Kerberos (not Negotiate) was used, we need to send a special protocol value and not SPNEGO.
                    protocol = WinRMEncryption.KERBEROS

                self.encryption = WinRMEncryption(self.session.auth.contexts[hostname], protocol)  # type: ignore[union-attr] # This should not happen

        if log.isEnabledFor(logging.DEBUG):
            log.debug("Sending message: %s" % message.decode("utf-8"))
        # for testing, keep commented out
        # self._test_messages.append({"request": message.decode('utf-8'),
        #                             "response": None})

        headers = self.session.headers
        if self.wrap_required:
            content_type, payload = self.encryption.wrap_message(message)  # type: ignore[union-attr] # This should not happen
            protocol = self.encryption.protocol if self.encryption else WinRMEncryption.SPNEGO
            type_header = '%s;protocol="%s";boundary="Encrypted Boundary"' % (content_type, protocol)
            headers.update(
                {
                    "Content-Type": type_header,
                    "Content-Length": str(len(payload)),
                }
            )
        else:
            payload = message
            headers["Content-Type"] = "application/soap+xml;charset=UTF-8"

        request = requests.Request("POST", self.endpoint, data=payload, headers=headers)
        prep_request = self.session.prepare_request(request)
        return self._send_request(prep_request)

    def _send_request(self, request: requests.PreparedRequest) -> bytes:
        response = self.session.send(request, timeout=(self.connection_timeout, self.read_timeout))  # type: ignore[union-attr] # This should not happen

        content_type = response.headers.get("content-type", "")
        if content_type.startswith("multipart/encrypted;") or content_type.startswith("multipart/x-multi-encrypted;"):
            boundary = re.search("boundary=[" '|\\"](.*)[' '|\\"]', response.headers["content-type"]).group(1)  # type: ignore[union-attr] # This should not happen
            response_content = self.encryption.unwrap_message(response.content, to_unicode(boundary))  # type: ignore[union-attr] # This should not happen
            response_text = to_string(response_content)
        else:
            response_content = response.content
            response_text = response.text if response_content else ""

        log.debug("Received message: %s" % response_text)
        # for testing, keep commented out
        # self._test_messages[-1]['response'] = response_text
        try:
            response.raise_for_status()
        except requests.HTTPError as err:
            response = err.response
            if response.status_code == 401:
                raise AuthenticationError("Failed to authenticate the user %s with %s" % (self.username, self.auth))
            else:
                code = response.status_code
                raise WinRMTransportError("http", code, response_text)

        return response_content

    def _build_session(self) -> requests.Session:
        log.debug("Building requests session with auth %s" % self.auth)
        self._suppress_library_warnings()

        session = requests.Session()
        session.headers["User-Agent"] = "Python PSRP Client"

        # requests defaults to 'Accept-Encoding: gzip, default' which normally doesn't matter on vanila WinRM but for
        # Exchange endpoints hosted on IIS they actually compress it with 1 of the 2 algorithms. By explicitly setting
        # identity we are telling the server not to transform (compress) the data using the HTTP methods which we don't
        # support. https://tools.ietf.org/html/rfc7231#section-5.3.4
        session.headers["Accept-Encoding"] = "identity"

        # get the env requests settings
        session.trust_env = True
        settings = session.merge_environment_settings(  # type: ignore[no-untyped-call] # Not in types-requests
            url=self.endpoint, proxies={}, stream=None, verify=None, cert=None
        )

        # set the proxy config
        session.proxies = settings["proxies"]
        proxy_key = "https" if self.ssl else "http"
        if self.proxy is not None:
            session.proxies = {
                proxy_key: self.proxy,
            }
        elif self.no_proxy:
            session.proxies = {
                proxy_key: False,  # type: ignore[dict-item] # A boolean is expected here
            }

        # Retry on connection errors, with a backoff factor
        retry_kwargs = {
            "total": self.reconnection_retries,
            "connect": self.reconnection_retries,
            "status": self.reconnection_retries,
            "read": 0,
            "backoff_factor": self.reconnection_backoff,
            "status_forcelist": (425, 429, 503),
        }
        try:
            retries = Retry(**retry_kwargs)
        except TypeError:
            # Status was added in urllib3 >= 1.21 (Requests >= 2.14.0), remove
            # the status retry counter and try again. The user should upgrade
            # to a newer version
            log.warning(
                "Using an older requests version that without support for status retries, ignoring.", exc_info=True
            )
            del retry_kwargs["status"]
            retries = Retry(**retry_kwargs)

        session.mount("http://", requests.adapters.HTTPAdapter(max_retries=retries))
        session.mount("https://", requests.adapters.HTTPAdapter(max_retries=retries))

        # set cert validation config
        session.verify = self.cert_validation

        # if cert_validation is a bool (no path specified), not False and there
        # are env settings for verification, set those env settings
        if isinstance(self.cert_validation, bool) and self.cert_validation and settings["verify"] is not None:
            session.verify = settings["verify"]

        build_auth = getattr(self, "_build_auth_%s" % self.auth)
        build_auth(session)
        return session

    def _build_auth_basic(self, session: requests.Session) -> None:
        if self.username is None:
            raise ValueError("For basic auth, the username must be specified")
        if self.password is None:
            raise ValueError("For basic auth, the password must be specified")

        session.auth = requests.auth.HTTPBasicAuth(username=self.username, password=self.password)

    def _build_auth_certificate(self, session: requests.Session) -> None:
        if self.certificate_key_pem is None:
            raise ValueError(
                "For certificate auth, the path to the "
                "certificate key pem file must be specified with "
                "certificate_key_pem"
            )
        if self.certificate_pem is None:
            raise ValueError(
                "For certificate auth, the path to the "
                "certificate pem file must be specified with "
                "certificate_pem"
            )
        if self.ssl is False:
            raise ValueError("For certificate auth, SSL must be used")

        session.cert = (self.certificate_pem, self.certificate_key_pem)
        session.headers["Authorization"] = "http://schemas.dmtf.org/wbem/wsman/1/wsman/secprofile/https/mutual"

    def _build_auth_credssp(self, session: requests.Session) -> None:
        if self.username is None:
            raise ValueError("For credssp auth, the username must be specified")
        if self.password is None:
            raise ValueError("For credssp auth, the password must be specified")

        kwargs = self._get_auth_kwargs("credssp")
        session.auth = HttpCredSSPAuth(username=self.username, password=self.password, **kwargs)

    def _build_auth_kerberos(self, session: requests.Session) -> None:
        self._build_auth_negotiate(session, "kerberos")

    def _build_auth_negotiate(self, session: requests.Session, auth_provider: str = "negotiate") -> None:
        kwargs = self._get_auth_kwargs("negotiate")

        session.auth = HTTPNegotiateAuth(
            username=self.username,
            password=self.password,
            auth_provider=auth_provider,
            wrap_required=self.wrap_required,
            **kwargs,
        )

    def _build_auth_ntlm(self, session: requests.Session) -> None:
        self._build_auth_negotiate(session, "ntlm")

    def _get_auth_kwargs(self, auth_provider: str) -> typing.Dict[str, typing.Any]:
        kwargs = {}
        for kwarg in AUTH_KWARGS[auth_provider]:
            kwarg_value = getattr(self, kwarg, None)
            if kwarg_value is not None:
                kwarg_key = kwarg[len(auth_provider) + 1 :]
                kwargs[kwarg_key] = kwarg_value

        return kwargs

    def _suppress_library_warnings(self) -> None:
        # try to suppress known warnings from requests if possible
        try:
            from requests.packages.urllib3.exceptions import InsecurePlatformWarning

            warnings.simplefilter("ignore", category=InsecurePlatformWarning)
        except:  # NOQA: E722; # pragma: no cover
            pass

        try:
            from requests.packages.urllib3.exceptions import SNIMissingWarning

            warnings.simplefilter("ignore", category=SNIMissingWarning)
        except:  # NOQA: E722; # pragma: no cover
            pass

        # if we're explicitly ignoring validation, try to suppress
        # InsecureRequestWarning, since the user opted-in
        if self.cert_validation is False:
            try:
                from requests.packages.urllib3.exceptions import InsecureRequestWarning

                warnings.simplefilter("ignore", category=InsecureRequestWarning)
            except:  # NOQA: E722; # pragma: no cover
                pass

    @staticmethod
    def _create_endpoint(
        ssl: bool,
        server: str,
        port: int,
        path: str,
    ) -> str:
        scheme = "https" if ssl else "http"

        # Check if the server is an IPv6 Address, enclose in [] if it is
        try:
            address = ipaddress.IPv6Address(to_unicode(server))
        except ipaddress.AddressValueError:
            pass
        else:
            server = "[%s]" % address.compressed

        endpoint = "%s://%s:%s/%s" % (scheme, server, port, path)
        return endpoint



# -*- coding: utf-8 -*-

from collections import OrderedDict

import pytest

from pypsrp.exceptions import WSManFaultError
from pypsrp.shell import Process, SignalCode, WinRS
from pypsrp.wsman import WSMan


class TestWinRS(object):
    @pytest.mark.parametrize("wsman_conn", [[True, "test_winrs_standard"]], indirect=True)
    def test_winrs_standard(self, wsman_conn):
        with WinRS(wsman_conn) as shell:
            process = Process(shell, "cmd.exe", ["/c", "echo", "hi"])
            process.invoke()
            process.signal(SignalCode.CTRL_C)
            assert process.rc == 0
            assert process.stdout == b"hi\r\n"
            assert process.stderr == b""

    @pytest.mark.parametrize(
        "wsman_conn",
        [
            [
                # no_shell only works on older hosts, rely on
                # pre-canned responses from actual test
                False,
                "test_winrs_no_cmd_shell",
            ]
        ],
        indirect=True,
    )
    def test_winrs_no_cmd_shell(self, wsman_conn):
        with WinRS(wsman_conn) as shell:
            process = Process(shell, "powershell.exe", ["Write-Host", "hi"], no_shell=True)

            # this will fail as you need to provide the full path when not
            # running in cmd shell
            with pytest.raises(WSManFaultError) as exc:
                process.invoke()
            assert exc.value.provider_fault == "The system cannot find the file specified."
            assert exc.value.code == 2147942402

            # fix the execute path and invoke again
            process.executable = r"C:\Windows\system32\WindowsPowerShell\v1.0\powershell.exe"
            process.invoke()
            process.signal(SignalCode.CTRL_C)

            assert process.rc == 0
            assert process.stdout == b"hi\n"
            assert process.stderr == b""

    @pytest.mark.parametrize(
        "wsman_conn",
        [
            [
                # to save on test runtime, only run with
                # pre-canned responses
                False,
                "test_winrs_operation_timeout",
            ]
        ],
        indirect=True,
    )
    def test_winrs_operation_timeout(self, wsman_conn):
        wsman_conn.operation_timeout = 10
        with WinRS(wsman_conn) as shell:
            process = Process(shell, "powershell.exe", ["Write-Host hi; Start-Sleep 30; Write-Host hi again"])
            process.invoke()
            process.signal(SignalCode.CTRL_C)
            assert process.rc == 0
            assert process.stdout == b"hi\nhi again\n"
            assert process.stderr == b""

    @pytest.mark.parametrize("wsman_conn", [[True, "test_winrs_stderr_rc"]], indirect=True)
    def test_winrs_stderr_rc(self, wsman_conn):
        with WinRS(wsman_conn) as shell:
            process = Process(shell, "cmd.exe", ["/c echo out && echo err>&2 && exit 1"])
            process.invoke()
            process.signal(SignalCode.CTRL_C)
            assert process.rc == 1
            assert process.stdout == b"out \r\n"
            assert process.stderr == b"err \r\n"

    @pytest.mark.parametrize("wsman_conn", [[True, "test_winrs_send"]], indirect=True)
    def test_winrs_send(self, wsman_conn):
        with WinRS(wsman_conn) as shell:
            process = Process(shell, "powershell.exe", ["-"])
            process.begin_invoke()
            process.send(b'Write-Host "output 1";', end=False)
            process.send(b'Write-Host "output 2";')
            process.end_invoke()
            process.signal(SignalCode.CTRL_C)
            assert process.rc == 0
            assert process.stdout == b"output 1\noutput 2\n"
            assert process.stderr == b""

    @pytest.mark.parametrize("wsman_conn", [[True, "test_winrs_environment"]], indirect=True)
    def test_winrs_environment(self, wsman_conn):
        complex_chars = r'_-(){}[]<>*+-/\?"' "!@#$^&|;:i,.`~0"
        env_block = OrderedDict(
            [
                ("env1", "var1"),
                (1234, 5678),
                (complex_chars, complex_chars),
            ]
        )

        with WinRS(wsman_conn, environment=env_block) as shell:
            process = Process(shell, "cmd.exe", ["/c", "set"])
            process.invoke()
            process.signal(SignalCode.CTRL_C)
            env_list = process.stdout.decode("utf-8").splitlines()
            assert process.rc == 0
            assert "env1=var1" in env_list
            assert "1234=5678" in env_list
            assert "%s=%s" % (complex_chars, complex_chars) in env_list
            assert process.stderr == b""

    @pytest.mark.parametrize("wsman_conn", [[True, "test_winrs_extra_opts"]], indirect=True)
    def test_winrs_extra_opts(self, wsman_conn):
        with WinRS(wsman_conn, name="shell 1", lifetime=60, idle_time_out=60, working_directory="C:\\Windows") as shell:
            assert shell.name == "shell 1"
            assert shell.lifetime == 60
            assert shell.idle_time_out == "PT60.000S"
            assert shell.working_directory == "C:\\Windows"

            process = Process(shell, "powershell.exe", ["(pwd).Path"])
            process.invoke()
            process.signal(SignalCode.CTRL_C)
            assert process.rc == 0
            assert process.stdout == b"C:\\Windows\r\n"
            assert process.stderr == b""

    @pytest.mark.parametrize("wsman_conn", [[True, "test_winrs_unicode"]], indirect=True)
    def test_winrs_unicode(self, wsman_conn):
        with WinRS(wsman_conn, codepage=65001) as shell:
            process = Process(shell, "powershell.exe", ["Write-Host こんにちは"])
            process.invoke()
            process.signal(SignalCode.CTRL_C)
            assert process.rc == 0
            assert process.stdout.decode("utf-8") == "こんにちは\n"
            assert process.stderr == b""

    @pytest.mark.parametrize(
        "wsman_conn",
        # not all hosts respect the no_profile, we will
        # just validate the message against a fake host
        [[False, "test_winrs_noprofile"]],
        indirect=True,
    )
    def test_winrs_noprofile(self, wsman_conn):
        with WinRS(wsman_conn, no_profile=True) as shell:
            process = Process(shell, "cmd.exe", ["/c", "set"])
            process.invoke()
            process.signal(SignalCode.CTRL_C)
            assert process.rc == 0
            assert "USERPROFILE=C:\\Users\\Default" in process.stdout.decode("utf-8").splitlines()
            assert process.stderr == b""

    @pytest.mark.parametrize("wsman_conn", [[True, "test_winrs_open_already_opened"]], indirect=True)
    def test_winrs_open_already_opened(self, wsman_conn):
        with WinRS(wsman_conn) as shell:
            shell.open()
        shell.close()

    @pytest.mark.parametrize("wsman_conn", [[True, "test_winrs_fail_poll_process"]], indirect=True)
    def test_winrs_fail_poll_process(self, wsman_conn):
        with WinRS(wsman_conn) as shell:
            process = Process(shell, "cmd.exe", ["/c", "echo", "hi"])

            # if I poll before beginning it should fail
            with pytest.raises(WSManFaultError) as err:
                process.poll_invoke()
            assert err.value.code == 87
            assert (
                err.value.message == "Received a WSManFault message. (Code: 87, Machine: {0}, "
                "Reason: The parameter is incorrect., Provider: Shell cmd "
                "plugin, Provider Path: %systemroot%\\system32\\winrscmd.dll"
                ", Provider Fault: The parameter is incorrect.)".format(err.value.machine)
            )
            assert err.value.provider == "Shell cmd plugin"
            assert err.value.provider_fault == "The parameter is incorrect."
            assert err.value.provider_path == "%systemroot%\\system32\\winrscmd.dll"
            assert err.value.reason == "The parameter is incorrect."



import base64
import re
import warnings

import pytest
import requests
from urllib3.response import HTTPResponse

from pypsrp.exceptions import AuthenticationError
from pypsrp.negotiate import HTTPNegotiateAuth


class TestTokenHelpers(object):
    def test_auth_supported(self):
        response = requests.Response()
        response.headers["www-authenticate"] = "Negotiate"
        HTTPNegotiateAuth._check_auth_supported(response, "Negotiate")

    def test_auth_supported_multiple(self):
        response = requests.Response()
        response.headers["www-authenticate"] = "Negotiate, CredSSP, Basic Realm='WSMan'"
        HTTPNegotiateAuth._check_auth_supported(response, "Negotiate")

    def test_auth_not_supported(self):
        response = requests.Response()
        response.headers["www-authenticate"] = "CredSSP"

        expected = (
            "The server did not response with one of the following authentication methods Negotiate - "
            "actual: 'CredSSP'"
        )
        with pytest.raises(AuthenticationError, match=re.escape(expected)):
            HTTPNegotiateAuth._check_auth_supported(response, ["Negotiate"])

    def test_auth_not_supported_no_header(self):
        response = requests.Response()

        expected = "The server did not response with one of the following authentication methods Negotiate - actual: ''"
        with pytest.raises(AuthenticationError, match=re.escape(expected)):
            HTTPNegotiateAuth._check_auth_supported(response, ["Negotiate"])

    def test_set_auth_token(self):
        request = requests.Request("GET", "")
        expected = b"Negotiate YWJj"
        HTTPNegotiateAuth._set_auth_token(request, b"abc", "Negotiate")
        actual = request.headers["Authorization"]
        assert actual == expected

    def test_get_auth_token_present(self):
        auth = HTTPNegotiateAuth()
        response = requests.Response()
        response.headers["www-authenticate"] = "Negotiate YWJj"
        expected = b"abc"
        actual = HTTPNegotiateAuth._get_auth_token(response, auth._regex)
        assert actual == expected

    def test_get_auth_token_no_header(self):
        auth = HTTPNegotiateAuth()
        response = requests.Response()
        expected = None
        actual = HTTPNegotiateAuth._get_auth_token(response, auth._regex)
        assert actual == expected

    @pytest.mark.parametrize("header", ["Kerberos", "Negotiate", "NTLM"])
    def test_get_auth_token_kerberos_auth(self, header):
        auth = HTTPNegotiateAuth()
        response = requests.Response()
        response.headers["www-authenticate"] = "%s YWJj" % header
        expected = b"abc"
        actual = HTTPNegotiateAuth._get_auth_token(response, auth._regex)
        assert actual == expected

    def test_get_auth_token_different_auth(self):
        auth = HTTPNegotiateAuth()
        response = requests.Response()
        response.headers["www-authenticate"] = "Fake YWJj"
        expected = None
        actual = HTTPNegotiateAuth._get_auth_token(response, auth._regex)
        assert actual == expected


class TestCertificateHash(object):
    def test_rsa_md5(self):
        cert_der = (
            b"MIIDGzCCAgOgAwIBAgIQJzshhViMG5hLHIJHxa+TcTANBgkqhkiG9w0"
            b"BAQQFADAVMRMwEQYDVQQDDApTRVJWRVIyMDE2MB4XDTE3MDUzMDA4MD"
            b"MxNloXDTE4MDUzMDA4MjMxNlowFTETMBEGA1UEAwwKU0VSVkVSMjAxN"
            b"jCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAN9N5GAzI7uq"
            b"AVlI6vUqhY5+EZWCWWGRwR3FT2DEXE5++AiJxXO0i0ZfAkLu7UggtBe"
            b"QwVNkaPD27EYzVUhy1iDo37BrFcLNpfjsjj8wVjaSmQmqvLvrvEh/BT"
            b"C5SBgDrk2+hiMh9PrpJoB3QAMDinz5aW0rEXMKitPBBiADrczyYrliF"
            b"AlEU6pTlKEKDUAeP7dKOBlDbCYvBxKnR3ddVH74I5T2SmNBq5gzkbKP"
            b"nlCXdHLZSh74USu93rKDZQF8YzdTO5dcBreJDJsntyj1o49w9WCt6M7"
            b"+pg6vKvE+tRbpCm7kXq5B9PDi42Nb6//MzNaMYf9V7v5MHapvVSv3+y"
            b"sCAwEAAaNnMGUwDgYDVR0PAQH/BAQDAgWgMB0GA1UdJQQWMBQGCCsGA"
            b"QUFBwMCBggrBgEFBQcDATAVBgNVHREEDjAMggpTRVJWRVIyMDE2MB0G"
            b"A1UdDgQWBBTh4L2Clr9ber6yfY3JFS3wiECL4DANBgkqhkiG9w0BAQQ"
            b"FAAOCAQEA0JK/SL7SP9/nvqWp52vnsxVefTFehThle5DLzagmms/9gu"
            b"oSE2I9XkQIttFMprPosaIZWt7WP42uGcZmoZOzU8kFFYJMfg9Ovyca+"
            b"gnG28jDUMF1E74KrC7uynJiQJ4vPy8ne7F3XJ592LsNJmK577l42gAW"
            b"u08p3TvEJFNHy2dBk/IwZp0HIPr9+JcPf7v0uL6lK930xHJHP56XLzN"
            b"YG8vCMpJFR7wVZp3rXkJQUy3GxyHPJPjS8S43I9j+PoyioWIMEotq2+"
            b"q0IpXU/KeNFkdGV6VPCmzhykijExOMwO6doUzIUM8orv9jYLHXYC+i6"
            b"IFKSb6runxF1MAik+GCSA=="
        )
        cert_der = base64.b64decode(cert_der)

        expected = (
            b"\x23\x34\xB8\x47\x6C\xBF\x4E\x6D\xFC\x76\x6A\x5D"
            b"\x5A\x30\xD6\x64\x9C\x01\xBA\xE1\x66\x2A\x5C\x3A"
            b"\x13\x02\xA9\x68\xD7\xC6\xB0\xF6"
        )
        actual = HTTPNegotiateAuth._get_certificate_hash(cert_der)
        assert actual == expected

    def test_rsa_sha1(self):
        cert_der = (
            b"MIIDGzCCAgOgAwIBAgIQJg/Mf5sR55xApJRK+kabbTANBgkqhkiG9w0"
            b"BAQUFADAVMRMwEQYDVQQDDApTRVJWRVIyMDE2MB4XDTE3MDUzMDA4MD"
            b"MxNloXDTE4MDUzMDA4MjMxNlowFTETMBEGA1UEAwwKU0VSVkVSMjAxN"
            b"jCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBALPKwYikjbzL"
            b"Lo6JtS6cyytdMMjSrggDoTnRUKauC5/izoYJd+2YVR5YqnluBJZpoFp"
            b"hkCgFFohUOU7qUsI1SkuGnjI8RmWTrrDsSy62BrfX+AXkoPlXo6IpHz"
            b"HaEPxjHJdUACpn8QVWTPmdAhwTwQkeUutrm3EOVnKPX4bafNYeAyj7/"
            b"AGEplgibuXT4/ehbzGKOkRN3ds/pZuf0xc4Q2+gtXn20tQIUt7t6iwh"
            b"nEWjIgopFL/hX/r5q5MpF6stc1XgIwJjEzqMp76w/HUQVqaYneU4qSG"
            b"f90ANK/TQ3aDbUNtMC/ULtIfHqHIW4POuBYXaWBsqalJL2VL3YYkKTU"
            b"sCAwEAAaNnMGUwDgYDVR0PAQH/BAQDAgWgMB0GA1UdJQQWMBQGCCsGA"
            b"QUFBwMCBggrBgEFBQcDATAVBgNVHREEDjAMggpTRVJWRVIyMDE2MB0G"
            b"A1UdDgQWBBS1jgojcjPu9vqeP1uSKuiIonGwAjANBgkqhkiG9w0BAQU"
            b"FAAOCAQEAKjHL6k5Dv/Zb7dvbYEZyx0wVhjHkCTpT3xstI3+TjfAFsu"
            b"3zMmyFqFqzmr4pWZ/rHc3ObD4pEa24kP9hfB8nmr8oHMLebGmvkzh5h"
            b"0GYc4dIH7Ky1yfQN51hi7/X5iN7jnnBoCJTTlgeBVYDOEBXhfXi3cLT"
            b"u3d7nz2heyNq07gFP8iN7MfqdPZndVDYY82imLgsgar9w5d+fvnYM+k"
            b"XWItNNCUH18M26Obp4Es/Qogo/E70uqkMHost2D+tww/7woXi36X3w/"
            b"D2yBDyrJMJKZLmDgfpNIeCimncTOzi2IhzqJiOY/4XPsVN/Xqv0/dzG"
            b"TDdI11kPLq4EiwxvPanCg=="
        )
        cert_der = base64.b64decode(cert_der)

        expected = (
            b"\x14\xCF\xE8\xE4\xB3\x32\xB2\x0A\x34\x3F\xC8\x40"
            b"\xB1\x8F\x9F\x6F\x78\x92\x6A\xFE\x7E\xC3\xE7\xB8"
            b"\xE2\x89\x69\x61\x9B\x1E\x8F\x3E"
        )
        actual = HTTPNegotiateAuth._get_certificate_hash(cert_der)
        assert actual == expected

    def test_rsa_sha256(self):
        cert_der = (
            b"MIIDGzCCAgOgAwIBAgIQWkeAtqoFg6pNWF7xC4YXhTANBgkqhkiG9w0"
            b"BAQsFADAVMRMwEQYDVQQDDApTRVJWRVIyMDE2MB4XDTE3MDUyNzA5MD"
            b"I0NFoXDTE4MDUyNzA5MjI0NFowFTETMBEGA1UEAwwKU0VSVkVSMjAxN"
            b"jCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBALIPKM5uykFy"
            b"NmVoLyvPSXGk15ZDqjYi3AbUxVFwCkVImqhefLATit3PkTUYFtAT+TC"
            b"AwK2E4lOu1XHM+Tmp2KIOnq2oUR8qMEvfxYThEf1MHxkctFljFssZ9N"
            b"vASDD4lzw8r0Bhl+E5PhR22Eu1Wago5bvIldojkwG+WBxPQv3ZR546L"
            b"MUZNaBXC0RhuGj5w83lbVz75qM98wvv1ekfZYAP7lrVyHxqCTPDomEU"
            b"I45tQQZHCZl5nRx1fPCyyYfcfqvFlLWD4Q3PZAbnw6mi0MiWJbGYKME"
            b"1XGicjqyn/zM9XKA1t/JzChS2bxf6rsyA9I7ibdRHUxsm1JgKry2jfW"
            b"0CAwEAAaNnMGUwDgYDVR0PAQH/BAQDAgWgMB0GA1UdJQQWMBQGCCsGA"
            b"QUFBwMCBggrBgEFBQcDATAVBgNVHREEDjAMggpTRVJWRVIyMDE2MB0G"
            b"A1UdDgQWBBQabLGWg1sn7AXPwYPyfE0ER921ZDANBgkqhkiG9w0BAQs"
            b"FAAOCAQEAnRohyl6ZmOsTWCtxOJx5A8yr//NweXKwWWmFQXRmCb4bMC"
            b"xhD4zqLDf5P6RotGV0I/SHvqz+pAtJuwmr+iyAF6WTzo3164LCfnQEu"
            b"psfrrfMkf3txgDwQkA0oPAw3HEwOnR+tzprw3Yg9x6UoZEhi4XqP9AX"
            b"R49jU92KrNXJcPlz5MbkzNo5t9nr2f8q39b5HBjaiBJxzdM1hxqsbfD"
            b"KirTYbkUgPlVOo/NDmopPPb8IX8ubj/XETZG2jixD0zahgcZ1vdr/iZ"
            b"+50WSXKN2TAKBO2fwoK+2/zIWrGRxJTARfQdF+fGKuj+AERIFNh88HW"
            b"xSDYjHQAaFMcfdUpa9GGQ=="
        )
        cert_der = base64.b64decode(cert_der)

        expected = (
            b"\x99\x6F\x3E\xEA\x81\x2C\x18\x70\xE3\x05\x49\xFF"
            b"\x9B\x86\xCD\x87\xA8\x90\xB6\xD8\xDF\xDF\x4A\x81"
            b"\xBE\xF9\x67\x59\x70\xDA\xDB\x26"
        )
        actual = HTTPNegotiateAuth._get_certificate_hash(cert_der)
        assert actual == expected

    def test_rsa_sha384(self):
        cert_der = (
            b"MIIDGzCCAgOgAwIBAgIQEmj1prSSQYRL2zYBEjsm5jANBgkqhkiG9w0"
            b"BAQwFADAVMRMwEQYDVQQDDApTRVJWRVIyMDE2MB4XDTE3MDUzMDA4MD"
            b"MxN1oXDTE4MDUzMDA4MjMxN1owFTETMBEGA1UEAwwKU0VSVkVSMjAxN"
            b"jCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAKsK5NvHi4xO"
            b"081fRLMmPqKsKaHvXgPRykLA0SmKxpGJHfTAZzxojHVeVwOm87IvQj2"
            b"JUh/yrRwSi5Oqrvqx29l2IC/qQt2xkAQsO51/EWkMQ5OSJsl1MN3NXW"
            b"eRTKVoUuJzBs8XLmeraxQcBPyyLhq+WpMl/Q4ZDn1FrUEZfxV0POXgU"
            b"dI3ApuQNRtJOb6iteBIoQyMlnof0RswBUnkiWCA/+/nzR0j33j47IfL"
            b"nkmU4RtqkBlO13f6+e1GZ4lEcQVI2yZq4Zgu5VVGAFU2lQZ3aEVMTu9"
            b"8HEqD6heyNp2on5G/K/DCrGWYCBiASjnX3wiSz0BYv8f3HhCgIyVKhJ"
            b"8CAwEAAaNnMGUwDgYDVR0PAQH/BAQDAgWgMB0GA1UdJQQWMBQGCCsGA"
            b"QUFBwMCBggrBgEFBQcDATAVBgNVHREEDjAMggpTRVJWRVIyMDE2MB0G"
            b"A1UdDgQWBBQS/SI61S2UE8xwSgHxbkCTpZXo4TANBgkqhkiG9w0BAQw"
            b"FAAOCAQEAMVV/WMXd9w4jtDfSrIsKaWKGtHtiMPpAJibXmSakBRwLOn"
            b"5ZGXL2bWI/Ac2J2Y7bSzs1im2ifwmEqwzzqnpVKShIkZmtij0LS0SEr"
            b"6Fw5IrK8tD6SH+lMMXUTvp4/lLQlgRCwOWxry/YhQSnuprx8IfSPvil"
            b"kwZ0Ysim4Aa+X5ojlhHpWB53edX+lFrmR1YWValBnQ5DvnDyFyLR6II"
            b"Ialp4vmkzI9e3/eOgSArksizAhpXpC9dxQBiHXdhredN0X+1BVzbgzV"
            b"hQBEwgnAIPa+B68oDILaV0V8hvxrP6jFM4IrKoGS1cq0B+Ns0zkG7ZA"
            b"2Q0W+3nVwSxIr6bd6hw7g=="
        )
        cert_der = base64.b64decode(cert_der)

        expected = (
            b"\x34\xF3\x03\xC9\x95\x28\x6F\x4B\x21\x4A\x9B\xA6"
            b"\x43\x5B\x69\xB5\x1E\xCF\x37\x58\xEA\xBC\x2A\x14"
            b"\xD7\xA4\x3F\xD2\x37\xDC\x2B\x1A\x1A\xD9\x11\x1C"
            b"\x5C\x96\x5E\x10\x75\x07\xCB\x41\x98\xC0\x9F\xEC"
        )
        actual = HTTPNegotiateAuth._get_certificate_hash(cert_der)
        assert actual == expected

    def test_rsa_sha512(self):
        cert_der = (
            b"MIIDGzCCAgOgAwIBAgIQUDHcKGevZohJV+TkIIYC1DANBgkqhkiG9w0"
            b"BAQ0FADAVMRMwEQYDVQQDDApTRVJWRVIyMDE2MB4XDTE3MDUzMDA4MD"
            b"MxN1oXDTE4MDUzMDA4MjMxN1owFTETMBEGA1UEAwwKU0VSVkVSMjAxN"
            b"jCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAKr9bo/XXvHt"
            b"D6Qnhb1wyLg9lDQxxe/enH49LQihtVTZMwGf2010h81QrRUe/bkHTvw"
            b"K22s2lqj3fUpGxtEbYFWLAHxv6IFnIKd+Zi1zaCPGfas9ekqCSj3vZQ"
            b"j7lCJVGUGuuqnSDvsed6g2Pz/g6mJUa+TzjxN+8wU5oj5YVUK+aing1"
            b"zPSA2MDCfx3+YzjxVwNoGixOz6Yx9ijT4pUsAYQAf1o9R+6W1/IpGgu"
            b"oax714QILT9heqIowwlHzlUZc1UAYs0/JA4CbDZaw9hlJyzMqe/aE46"
            b"efqPDOpO3vCpOSRcSyzh02WijPvEEaPejQRWg8RX93othZ615MT7dqp"
            b"ECAwEAAaNnMGUwDgYDVR0PAQH/BAQDAgWgMB0GA1UdJQQWMBQGCCsGA"
            b"QUFBwMCBggrBgEFBQcDATAVBgNVHREEDjAMggpTRVJWRVIyMDE2MB0G"
            b"A1UdDgQWBBTgod3R6vejt6kOASAApA19xIG6kTANBgkqhkiG9w0BAQ0"
            b"FAAOCAQEAVfz0okK2bh3OQE8cWNbJ5PjJRSAJEqVUvYaTlS0Nqkyuaj"
            b"gicP3hb/pF8FvaVaB6r7LqgBxyW5NNL1xwdNLt60M2zaULL6Fhm1vzM"
            b"sSMc2ynkyN4++ODwii674YcQAnkUh+ZGIx+CTdZBWJfVM9dZb7QjgBT"
            b"nVukeFwN2EOOBSpiQSBpcoeJEEAq9csDVRhEfcB8Wtz7TTItgOVsilY"
            b"dQY56ON5XszjCki6UA3GwdQbBEHjWF2WERqXWrojrSSNOYDvxM5mrEx"
            b"sG1npzUTsaIr9w8ty1beh/2aToCMREvpiPFOXnVV/ovHMU1lFQTNeQ0"
            b"OI7elR0nJ0peai30eMpQQ=="
        )
        cert_der = base64.b64decode(cert_der)

        expected = (
            b"\x55\x6E\x1C\x17\x84\xE3\xB9\x57\x37\x0B\x7F\x54"
            b"\x4F\x62\xC5\x33\xCB\x2C\xA5\xC1\xDA\xE0\x70\x6F"
            b"\xAE\xF0\x05\x44\xE1\xAD\x2B\x76\xFF\x25\xCF\xBE"
            b"\x69\xB1\xC4\xE6\x30\xC3\xBB\x02\x07\xDF\x11\x31"
            b"\x4C\x67\x38\xBC\xAE\xD7\xE0\x71\xD7\xBF\xBF\x2C"
            b"\x9D\xFA\xB8\x5D"
        )
        actual = HTTPNegotiateAuth._get_certificate_hash(cert_der)
        assert actual == expected

    def test_ecdsa_sha1(self):
        cert_der = (
            b"MIIBjjCCATSgAwIBAgIQRCJw7nbtvJ5F8wikRmwgizAJBgcqhkjOPQQ"
            b"BMBUxEzARBgNVBAMMClNFUlZFUjIwMTYwHhcNMTcwNTMwMDgwMzE3Wh"
            b"cNMTgwNTMwMDgyMzE3WjAVMRMwEQYDVQQDDApTRVJWRVIyMDE2MFkwE"
            b"wYHKoZIzj0CAQYIKoZIzj0DAQcDQgAEk3fOh178kRglmnPKe9K/mbgi"
            b"gf8YgNq62rF2EpfzpyQY0eGw4xnmKDG73aZ+ATSlV2IybxiUVsKyMUn"
            b"LhPfvmaNnMGUwDgYDVR0PAQH/BAQDAgWgMB0GA1UdJQQWMBQGCCsGAQ"
            b"UFBwMCBggrBgEFBQcDATAVBgNVHREEDjAMggpTRVJWRVIyMDE2MB0GA"
            b"1UdDgQWBBQSK8qwmiQmyAWWya3FxQDj9wqQAzAJBgcqhkjOPQQBA0kA"
            b"MEYCIQCiOsP56Iqo+cHRvCp2toj65Mgxo/PQY1tn+S3WH4RJFQIhAJe"
            b"gGQuaPWg6aCWV+2+6pNCNMdg/Nix+mMOJ88qCBNHi"
        )
        cert_der = base64.b64decode(cert_der)

        expected = (
            b"\x1E\xC9\xAD\x46\xDE\xE9\x34\x0E\x45\x03\xCF\xFD"
            b"\xB5\xCD\x81\x0C\xB2\x6B\x77\x8F\x46\xBE\x95\xD5"
            b"\xEA\xF9\x99\xDC\xB1\xC4\x5E\xDA"
        )
        actual = HTTPNegotiateAuth._get_certificate_hash(cert_der)
        assert actual == expected

    def test_ecdsa_sha256(self):
        cert_der = (
            b"MIIBjzCCATWgAwIBAgIQeNQTxkMgq4BF9tKogIGXUTAKBggqhkjOPQQ"
            b"DAjAVMRMwEQYDVQQDDApTRVJWRVIyMDE2MB4XDTE3MDUzMDA4MDMxN1"
            b"oXDTE4MDUzMDA4MjMxN1owFTETMBEGA1UEAwwKU0VSVkVSMjAxNjBZM"
            b"BMGByqGSM49AgEGCCqGSM49AwEHA0IABDAfXTLOaC3ElgErlgk2tBlM"
            b"wf9XmGlGBw4vBtMJap1hAqbsdxFm6rhK3QU8PFFpv8Z/AtRG7ba3UwQ"
            b"prkssClejZzBlMA4GA1UdDwEB/wQEAwIFoDAdBgNVHSUEFjAUBggrBg"
            b"EFBQcDAgYIKwYBBQUHAwEwFQYDVR0RBA4wDIIKU0VSVkVSMjAxNjAdB"
            b"gNVHQ4EFgQUnFDE8824TYAiBeX4fghEEg33UgYwCgYIKoZIzj0EAwID"
            b"SAAwRQIhAK3rXA4/0i6nm/U7bi6y618Ci2Is8++M3tYIXnEsA7zSAiA"
            b"w2s6bJoI+D7Xaey0Hp0gkks9z55y976keIEI+n3qkzw=="
        )
        cert_der = base64.b64decode(cert_der)

        expected = (
            b"\xFE\xCF\x1B\x25\x85\x44\x99\x90\xD9\xE3\xB2\xC9"
            b"\x2D\x3F\x59\x7E\xC8\x35\x4E\x12\x4E\xDA\x75\x1D"
            b"\x94\x83\x7C\x2C\x89\xA2\xC1\x55"
        )
        actual = HTTPNegotiateAuth._get_certificate_hash(cert_der)
        assert actual == expected

    def test_ecdsa_sha384(self):
        cert_der = (
            b"MIIBjzCCATWgAwIBAgIQcO3/jALdQ6BOAoaoseLSCjAKBggqhkjOPQQ"
            b"DAzAVMRMwEQYDVQQDDApTRVJWRVIyMDE2MB4XDTE3MDUzMDA4MDMxOF"
            b"oXDTE4MDUzMDA4MjMxOFowFTETMBEGA1UEAwwKU0VSVkVSMjAxNjBZM"
            b"BMGByqGSM49AgEGCCqGSM49AwEHA0IABJLjZH274heB/8PhmhWWCIVQ"
            b"Wle1hBZEN3Tk2yWSKaz9pz1bjwb9t79lVpQE9tvGL0zP9AqJYHcVOO9"
            b"YG9trqfejZzBlMA4GA1UdDwEB/wQEAwIFoDAdBgNVHSUEFjAUBggrBg"
            b"EFBQcDAgYIKwYBBQUHAwEwFQYDVR0RBA4wDIIKU0VSVkVSMjAxNjAdB"
            b"gNVHQ4EFgQUkRajoFr8qZ/8L8rKB3zGiGolDygwCgYIKoZIzj0EAwMD"
            b"SAAwRQIgfi8dAxXljCMSvngtDtagGCTGBs7Xxh8Z3WX6ZwJZsHYCIQC"
            b"D4iNReh1afXKYC0ipjXWAIkiihnEEycCIQMbkMNst7A=="
        )
        cert_der = base64.b64decode(cert_der)

        expected = (
            b"\xD2\x98\x7A\xD8\xF2\x0E\x83\x16\xA8\x31\x26\x1B"
            b"\x74\xEF\x7B\x3E\x55\x15\x5D\x09\x22\xE0\x7F\xFE"
            b"\x54\x62\x08\x06\x98\x2B\x68\xA7\x3A\x5E\x3C\x47"
            b"\x8B\xAA\x5E\x77\x14\x13\x5C\xB2\x6D\x98\x07\x49"
        )
        actual = HTTPNegotiateAuth._get_certificate_hash(cert_der)
        assert actual == expected

    def test_ecdsa_sha512(self):
        cert_der = (
            b"MIIBjjCCATWgAwIBAgIQHVj2AGEwd6pOOSbcf0skQDAKBggqhkjOPQQ"
            b"DBDAVMRMwEQYDVQQDDApTRVJWRVIyMDE2MB4XDTE3MDUzMDA3NTUzOV"
            b"oXDTE4MDUzMDA4MTUzOVowFTETMBEGA1UEAwwKU0VSVkVSMjAxNjBZM"
            b"BMGByqGSM49AgEGCCqGSM49AwEHA0IABL8d9S++MFpfzeH8B3vG/PjA"
            b"AWg8tGJVgsMw9nR+OfC9ltbTUwhB+yPk3JPcfW/bqsyeUgq4//LhaSp"
            b"lOWFNaNqjZzBlMA4GA1UdDwEB/wQEAwIFoDAdBgNVHSUEFjAUBggrBg"
            b"EFBQcDAgYIKwYBBQUHAwEwFQYDVR0RBA4wDIIKU0VSVkVSMjAxNjAdB"
            b"gNVHQ4EFgQUKUkCgLlxoeai0EtQrZth1/BSc5kwCgYIKoZIzj0EAwQD"
            b"RwAwRAIgRrV7CLpDG7KueyFA3ZDced9dPOcv2Eydx/hgrfxYEcYCIBQ"
            b"D35JvzmqU05kSFV5eTvkhkaDObd7V55vokhm31+Li"
        )
        cert_der = base64.b64decode(cert_der)

        expected = (
            b"\xE5\xCB\x68\xB2\xF8\x43\xD6\x3B\xF4\x0B\xCB\x20"
            b"\x07\x60\x8F\x81\x97\x61\x83\x92\x78\x3F\x23\x30"
            b"\xE5\xEF\x19\xA5\xBD\x8F\x0B\x2F\xAA\xC8\x61\x85"
            b"\x5F\xBB\x63\xA2\x21\xCC\x46\xFC\x1E\x22\x6A\x07"
            b"\x24\x11\xAF\x17\x5D\xDE\x47\x92\x81\xE0\x06\x87"
            b"\x8B\x34\x80\x59"
        )
        actual = HTTPNegotiateAuth._get_certificate_hash(cert_der)
        assert actual == expected

    def test_invalid_signature_algorithm(self):
        # Manually edited from test_ecdsa_sha512 to change the OID to
        # '1.2.840.10045.4.3.5'
        cert_der = (
            b"MIIBjjCCATWgAwIBAgIQHVj2AGEwd6pOOSbcf0skQDAKBggqhkjOPQQ"
            b"DBTAVMRMwEQYDVQQDDApTRVJWRVIyMDE2MB4XDTE3MDUzMDA3NTUzOV"
            b"oXDTE4MDUzMDA4MTUzOVowFTETMBEGA1UEAwwKU0VSVkVSMjAxNjBZM"
            b"BMGByqGSM49AgEGCCqGSM49AwEHA0IABL8d9S++MFpfzeH8B3vG/PjA"
            b"AWg8tGJVgsMw9nR+OfC9ltbTUwhB+yPk3JPcfW/bqsyeUgq4//LhaSp"
            b"lOWFNaNqjZzBlMA4GA1UdDwEB/wQEAwIFoDAdBgNVHSUEFjAUBggrBg"
            b"EFBQcDAgYIKwYBBQUHAwEwFQYDVR0RBA4wDIIKU0VSVkVSMjAxNjAdB"
            b"gNVHQ4EFgQUKUkCgLlxoeai0EtQrZth1/BSc5kwCgYIKoZIzj0EAwUD"
            b"RwAwRAIgRrV7CLpDG7KueyFA3ZDced9dPOcv2Eydx/hgrfxYEcYCIBQ"
            b"D35JvzmqU05kSFV5eTvkhkaDObd7V55vokhm31+Li"
        )
        cert_der = base64.b64decode(cert_der)

        expected = (
            b"\x65\xE1\xC7\x51\xAC\x33\xE0\x68\x03\xC3\xC9\xC2\x23\x45\x48\x43"
            b"\x43\x25\x45\xD6\x4B\x49\x25\xF3\xAE\xB2\xDD\xE5\x9B\x79\xF4\x39"
        )
        expected_warning = "Failed to get the signature algorithm from the certificate due to:"

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            actual = HTTPNegotiateAuth._get_certificate_hash(cert_der)
            assert actual == expected
            assert expected_warning in str(w[-1].message)

    def test_get_cbt_fail_not_urllib3(self, mocker):
        response = mocker.MagicMock()

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            actual = HTTPNegotiateAuth._get_cbt_data(response)
            assert actual is None
            assert (
                str(w[-1].message) == "Requests is running with a non urllib3 backend, cannot "
                "retrieve server cert for CBT. Raw type: MagicMock"
            )

    def test_get_cbt_fail_invalid_raw_socket(self, mocker):
        response = mocker.MagicMock()
        response.raw = HTTPResponse()

        expected_warning = "Failed to get raw socket for CBT from urllib3 resp"

        with warnings.catch_warnings(record=True) as w:
            actual = HTTPNegotiateAuth._get_cbt_data(response)
            assert actual is None
            assert expected_warning in str(w[-1].message)

    def test_get_cbt_no_peer_cert(self, mocker):
        mock_socket = mocker.MagicMock()
        mock_socket.getpeercert.side_effect = AttributeError

        raw_response = HTTPResponse()
        raw_response._fp = mocker.MagicMock()
        raw_response._fp.fp.raw._sock = mock_socket
        raw_response._fp.fp._sock = mock_socket

        response = mocker.MagicMock()
        response.raw = raw_response

        actual = HTTPNegotiateAuth._get_cbt_data(response)
        assert actual is None

    def test_get_cbt_with_peer_cert(self, mocker):
        cert_der = (
            b"MIIDGzCCAgOgAwIBAgIQWkeAtqoFg6pNWF7xC4YXhTANBgkqhkiG9w0"
            b"BAQsFADAVMRMwEQYDVQQDDApTRVJWRVIyMDE2MB4XDTE3MDUyNzA5MD"
            b"I0NFoXDTE4MDUyNzA5MjI0NFowFTETMBEGA1UEAwwKU0VSVkVSMjAxN"
            b"jCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBALIPKM5uykFy"
            b"NmVoLyvPSXGk15ZDqjYi3AbUxVFwCkVImqhefLATit3PkTUYFtAT+TC"
            b"AwK2E4lOu1XHM+Tmp2KIOnq2oUR8qMEvfxYThEf1MHxkctFljFssZ9N"
            b"vASDD4lzw8r0Bhl+E5PhR22Eu1Wago5bvIldojkwG+WBxPQv3ZR546L"
            b"MUZNaBXC0RhuGj5w83lbVz75qM98wvv1ekfZYAP7lrVyHxqCTPDomEU"
            b"I45tQQZHCZl5nRx1fPCyyYfcfqvFlLWD4Q3PZAbnw6mi0MiWJbGYKME"
            b"1XGicjqyn/zM9XKA1t/JzChS2bxf6rsyA9I7ibdRHUxsm1JgKry2jfW"
            b"0CAwEAAaNnMGUwDgYDVR0PAQH/BAQDAgWgMB0GA1UdJQQWMBQGCCsGA"
            b"QUFBwMCBggrBgEFBQcDATAVBgNVHREEDjAMggpTRVJWRVIyMDE2MB0G"
            b"A1UdDgQWBBQabLGWg1sn7AXPwYPyfE0ER921ZDANBgkqhkiG9w0BAQs"
            b"FAAOCAQEAnRohyl6ZmOsTWCtxOJx5A8yr//NweXKwWWmFQXRmCb4bMC"
            b"xhD4zqLDf5P6RotGV0I/SHvqz+pAtJuwmr+iyAF6WTzo3164LCfnQEu"
            b"psfrrfMkf3txgDwQkA0oPAw3HEwOnR+tzprw3Yg9x6UoZEhi4XqP9AX"
            b"R49jU92KrNXJcPlz5MbkzNo5t9nr2f8q39b5HBjaiBJxzdM1hxqsbfD"
            b"KirTYbkUgPlVOo/NDmopPPb8IX8ubj/XETZG2jixD0zahgcZ1vdr/iZ"
            b"+50WSXKN2TAKBO2fwoK+2/zIWrGRxJTARfQdF+fGKuj+AERIFNh88HW"
            b"xSDYjHQAaFMcfdUpa9GGQ=="
        )
        cert_der = base64.b64decode(cert_der)

        mock_socket = mocker.MagicMock()
        mock_socket.getpeercert.return_value = cert_der

        raw_response = HTTPResponse()
        raw_response._fp = mocker.MagicMock()
        raw_response._fp.fp.raw._sock = mock_socket
        raw_response._fp.fp._sock = mock_socket

        response = mocker.MagicMock()
        response.raw = raw_response

        expected = (
            b"tls-server-end-point:"
            b"\x99\x6F\x3E\xEA\x81\x2C\x18\x70\xE3\x05\x49\xFF"
            b"\x9B\x86\xCD\x87\xA8\x90\xB6\xD8\xDF\xDF\x4A\x81"
            b"\xBE\xF9\x67\x59\x70\xDA\xDB\x26"
        )
        actual = HTTPNegotiateAuth._get_cbt_data(response)
        assert actual == expected



# Copyright: (c) 2018, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import pkgutil
import typing
from urllib.parse import urlparse


def to_bytes(
    obj: typing.Any,
    encoding: str = "utf-8",
) -> bytes:
    """
    Makes sure the string is encoded as a byte string.

    :param obj: Python 2 string, Python 3 byte string, Unicode string to encode
    :param encoding: The encoding to use
    :return: The byte string that was encoded
    """
    if isinstance(obj, bytes):
        return obj

    return obj.encode(encoding)


def to_unicode(
    obj: typing.Any,
    encoding: str = "utf-8",
) -> str:
    """
    Makes sure the string is unicode string.

    :param obj: Python 2 string, Python 3 byte string, Unicode string to decode
    :param encoding: The encoding to use
    :return: THe unicode string the was decoded
    """
    if obj is None:
        obj = str(None)

    if isinstance(obj, str):
        return obj

    return obj.decode(encoding)


"""
Python 2 and 3 handle native strings differently, 2 is like a byte string while
3 uses unicode as the native string. The function to_string is used to easily
convert an existing string like object to the native version that is required
"""
to_string = to_unicode


def version_equal_or_newer(
    version: str,
    reference_version: str,
) -> bool:
    """
    Compares the 2 version strings and returns a bool that states whether
    version is newer than or equal to the reference version.

    This is quite strict and splits the string by . and compares the int
    values in them

    :param version: The version string to compare
    :param reference_version: The version string to check version against
    :return: True if version is newer than or equal to reference_version
    """
    version_parts = version.split(".")
    reference_version_parts = reference_version.split(".")

    # pad the version parts by 0 so the comparisons after won't fail with an
    # index error
    if len(version_parts) < len(reference_version_parts):
        diff = len(reference_version_parts) - len(version_parts)
        version_parts.extend(["0"] * diff)
    if len(reference_version_parts) < len(version_parts):
        diff = len(version_parts) - len(reference_version_parts)
        reference_version_parts.extend(["0"] * diff)

    newer = True
    for idx, version in enumerate(version_parts):
        current_version = int(reference_version_parts[idx])
        if int(version) < current_version:
            newer = False
            break
        elif int(version) > current_version:
            break

    return newer


def get_hostname(url: str) -> typing.Optional[str]:
    return urlparse(url).hostname


def get_pwsh_script(name: str) -> str:
    """
    Get the contents of a script stored in pypsrp/pwsh_scripts. Will also strip out any empty lines and comments to
    reduce the data we send across as much as possible.

    :param name: The filename of the script in pypsrp/pwsh_scripts to get.
    :return: The script contents.
    """
    script = to_unicode(pkgutil.get_data("pypsrp.pwsh_scripts", name))

    block_comment = False
    new_lines = []
    for line in script.splitlines():

        line = line.strip()
        if block_comment:
            block_comment = not line.endswith("#>")
        elif line.startswith("<#"):
            block_comment = True
        elif line and not line.startswith("#"):
            new_lines.append(line)

    return "\n".join(new_lines)



import xml.etree.ElementTree as ET

import pytest

from pypsrp._utils import to_unicode
from pypsrp.complex_objects import (
    Array,
    BufferCell,
    BufferCellType,
    Color,
    Command,
    CommandParameter,
    CommandType,
    Coordinates,
    HostInfo,
    ObjectMeta,
    Pipeline,
    PipelineResultTypes,
    PSThreadOptions,
    RemoteStreamOptions,
    Size,
)
from pypsrp.host import PSHost, PSHostRawUserInterface, PSHostUserInterface
from pypsrp.serializer import Serializer

from . import assert_xml_diff


def normalise_xml(xml_string):
    xml = "".join([l.lstrip() for l in to_unicode(xml_string).splitlines()])
    xml = ET.fromstring(xml)
    return to_unicode(ET.tostring(xml))


class TestEnum(object):
    def test_enum_invalid_value(self):
        state = PSThreadOptions(value=PSThreadOptions.DEFAULT)
        assert str(state) == "Default"
        state.value = 15
        with pytest.raises(KeyError) as err:
            str(state)
        assert "15 is not a valid enum value for System.Management.Automation.Runspaces.PSThreadOptions" in str(
            err.value
        )


class TestHostInfo(object):

    HOST_XML = """<Obj RefId="0">
            <MS>
                <Obj N="_hostDefaultData" RefId="1">
                    <MS>
                        <Obj N="data" RefId="2">
                            <TN RefId="0">
                                <T>System.Collections.Hashtable</T>
                                <T>System.Object</T>
                            </TN>
                            <DCT>
                                <En>
                                    <I32 N="Key">0</I32>
                                    <Obj N="Value" RefId="3">
                                        <MS>
                                            <S N="T">System.ConsoleColor</S>
                                            <I32 N="V">11</I32>
                                        </MS>
                                    </Obj>
                                </En>
                                <En>
                                    <I32 N="Key">1</I32>
                                    <Obj N="Value" RefId="4">
                                        <MS>
                                            <S N="T">System.ConsoleColor</S>
                                            <I32 N="V">12</I32>
                                        </MS>
                                    </Obj>
                                </En>
                                <En>
                                    <I32 N="Key">2</I32>
                                    <Obj N="Value" RefId="5">
                                        <MS>
                                            <S N="T">System.Management.Automation.Host.Coordinates</S>
                                            <Obj N="V" RefId="6">
                                                <MS>
                                                    <I32 N="x">1</I32>
                                                    <I32 N="y">2</I32>
                                                </MS>
                                            </Obj>
                                        </MS>
                                    </Obj>
                                </En>
                                <En>
                                    <I32 N="Key">3</I32>
                                    <Obj N="Value" RefId="7">
                                        <MS>
                                            <S N="T">System.Management.Automation.Host.Coordinates</S>
                                            <Obj N="V" RefId="8">
                                                <MS>
                                                    <I32 N="x">3</I32>
                                                    <I32 N="y">4</I32>
                                                </MS>
                                            </Obj>
                                        </MS>
                                    </Obj>
                                </En>
                                <En>
                                    <I32 N="Key">4</I32>
                                    <Obj N="Value" RefId="9">
                                        <MS>
                                            <S N="T">System.Int32</S>
                                            <I32 N="V">10</I32>
                                        </MS>
                                    </Obj>
                                </En>
                                <En>
                                    <I32 N="Key">5</I32>
                                    <Obj N="Value" RefId="10">
                                        <MS>
                                            <S N="T">System.Management.Automation.Host.Size</S>
                                            <Obj N="V" RefId="11">
                                                <MS>
                                                    <I32 N="height">10</I32>
                                                    <I32 N="width">20</I32>
                                                </MS>
                                            </Obj>
                                        </MS>
                                    </Obj>
                                </En>
                                <En>
                                    <I32 N="Key">6</I32>
                                    <Obj N="Value" RefId="12">
                                        <MS>
                                            <S N="T">System.Management.Automation.Host.Size</S>
                                            <Obj N="V" RefId="13">
                                                <MS>
                                                    <I32 N="height">30</I32>
                                                    <I32 N="width">40</I32>
                                                </MS>
                                            </Obj>
                                        </MS>
                                    </Obj>
                                </En>
                                <En>
                                    <I32 N="Key">7</I32>
                                    <Obj N="Value" RefId="14">
                                        <MS>
                                            <S N="T">System.Management.Automation.Host.Size</S>
                                            <Obj N="V" RefId="15">
                                                <MS>
                                                    <I32 N="height">50</I32>
                                                    <I32 N="width">60</I32>
                                                </MS>
                                            </Obj>
                                        </MS>
                                    </Obj>
                                </En>
                                <En>
                                    <I32 N="Key">8</I32>
                                    <Obj N="Value" RefId="16">
                                        <MS>
                                            <S N="T">System.Management.Automation.Host.Size</S>
                                            <Obj N="V" RefId="17">
                                                <MS>
                                                    <I32 N="height">70</I32>
                                                    <I32 N="width">80</I32>
                                                </MS>
                                            </Obj>
                                        </MS>
                                    </Obj>
                                </En>
                                <En>
                                    <I32 N="Key">9</I32>
                                    <Obj N="Value" RefId="18">
                                        <MS>
                                            <S N="T">System.String</S>
                                            <S N="V">Random Window Title</S>
                                        </MS>
                                    </Obj>
                                </En>
                            </DCT>
                        </Obj>
                    </MS>
                </Obj>
                <B N="_isHostNull">false</B>
                <B N="_isHostUINull">false</B>
                <B N="_isHostRawUINull">false</B>
                <B N="_useRunspaceHost">false</B>
            </MS>
        </Obj>"""

    def test_create_host_info(self):
        serializer = Serializer()

        foreground_color = Color(value=Color.CYAN)
        background_color = Color(value=Color.RED)
        cursor_position = Coordinates(x=1, y=2)
        window_position = Coordinates(x=3, y=4)
        cursor_size = 10
        buffer_size = Size(height=10, width=20)
        window_size = Size(height=30, width=40)
        max_window_size = Size(height=50, width=60)
        max_physical_window_size = Size(height=70, width=80)
        window_title = "Random Window Title"

        ps_raw_ui = PSHostRawUserInterface(
            window_title,
            cursor_size,
            foreground_color,
            background_color,
            cursor_position,
            window_position,
            buffer_size,
            max_physical_window_size,
            max_window_size,
            window_size,
        )
        ps_ui = PSHostUserInterface(raw_ui=ps_raw_ui)
        ps_host = PSHost(None, None, False, None, None, ps_ui, None)

        host_info = HostInfo(host=ps_host)
        expected_xml = normalise_xml(self.HOST_XML)

        actual = serializer.serialize(host_info)
        actual_xml = normalise_xml(ET.tostring(actual))
        assert_xml_diff(actual_xml, expected_xml)


class TestRemoteStreamOptions(object):
    def test_to_string_one_value(self):
        options = RemoteStreamOptions(value=1)
        expected = "AddInvocationInfoToErrorRecord"
        actual = str(options)
        assert actual == expected

    def test_to_string_multiple_values(self):
        options = RemoteStreamOptions(value=3)
        expected = "AddInvocationInfoToErrorRecord, AddInvocationInfoToWarningRecord"
        actual = str(options)
        assert actual == expected


class TestPipeline(object):

    PIPE_SINGLE = """<Obj RefId="0">
        <MS>
            <B N="IsNested">false</B>
            <Nil N="ExtraCmds"/>
            <Obj N="Cmds" RefId="1">
                <TN RefId="0">
                    <T>System.Collections.Generic.List`1[[System.Management.Automation.PSObject, System.Management.Automation, Version=1.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35]]</T>
                    <T>System.Object</T>
                </TN>
                <LST>
                    <Obj RefId="2">
                        <MS>
                            <S N="Cmd">Set-Variable</S>
                            <B N="IsScript">false</B>
                            <B N="UseLocalScope">false</B>
                            <Obj N="MergeMyResult" RefId="3">
                                <TN RefId="1">
                                    <T>System.Management.Automation.Runspaces.PipelineResultTypes</T>
                                    <T>System.Enum</T>
                                    <T>System.ValueType</T>
                                    <T>System.Object</T>
                                </TN>
                                <ToString>None</ToString>
                                <I32>0</I32>
                            </Obj>
                            <Obj N="MergeToResult" RefId="4">
                                <TNRef RefId="1"/>
                                <ToString>None</ToString>
                                <I32>0</I32>
                            </Obj>
                            <Obj N="MergePreviousResults" RefId="5">
                                <TNRef RefId="1"/>
                                <ToString>None</ToString>
                                <I32>0</I32>
                            </Obj>
                            <Obj N="Args" RefId="6">
                                <TNRef RefId="0"/>
                                <LST>
                                    <Obj RefId="7">
                                        <MS>
                                            <S N="N">Name</S>
                                            <S N="V">var</S>
                                        </MS>
                                    </Obj>
                                    <Obj RefId="8">
                                        <MS>
                                            <S N="N">Value</S>
                                            <S N="V">abc</S>
                                        </MS>
                                    </Obj>
                                </LST>
                            </Obj>
                            <Obj N="MergeError" RefId="9">
                                <TNRef RefId="1"/>
                                <ToString>None</ToString>
                                <I32>0</I32>
                            </Obj>
                            <Obj N="MergeWarning" RefId="10">
                                <TNRef RefId="1"/>
                                <ToString>None</ToString>
                                <I32>0</I32>
                            </Obj>
                            <Obj N="MergeVerbose" RefId="11">
                                <TNRef RefId="1"/>
                                <ToString>None</ToString>
                                <I32>0</I32>
                            </Obj>
                            <Obj N="MergeDebug" RefId="12">
                                <TNRef RefId="1"/>
                                <ToString>None</ToString>
                                <I32>0</I32>
                            </Obj>
                        </MS>
                    </Obj>
                </LST>
            </Obj>
            <Nil N="History"/>
            <B N="RedirectShellErrorOutputPipe">false</B>
        </MS>
    </Obj>"""

    PIPE_MULTIPLE = """<Obj RefId="0">
        <MS>
            <B N="IsNested">false</B>
            <Obj N="ExtraCmds" RefId="1">
                <TN RefId="0">
                    <T>System.Collections.Generic.List`1[[System.Management.Automation.PSObject, System.Management.Automation, Version=1.0.0.0, Culture=neutral, PublicKeyToken=31bf3856ad364e35]]</T>
                    <T>System.Object</T>
                </TN>
                <LST>
                    <Obj RefId="2">
                        <MS>
                            <Obj N="Cmds" RefId="3">
                                <TNRef RefId="0"/>
                                <LST>
                                    <Obj RefId="4">
                                        <MS>
                                            <S N="Cmd">Set-Variable</S>
                                            <B N="IsScript">false</B>
                                            <B N="UseLocalScope">false</B>
                                            <Obj N="MergeMyResult" RefId="5">
                                                <TN RefId="1">
                                                    <T>System.Management.Automation.Runspaces.PipelineResultTypes</T>
                                                    <T>System.Enum</T>
                                                    <T>System.ValueType</T>
                                                    <T>System.Object</T>
                                                </TN>
                                                <ToString>None</ToString>
                                                <I32>0</I32>
                                            </Obj>
                                            <Obj N="MergeToResult" RefId="6">
                                                <TNRef RefId="1"/>
                                                <ToString>None</ToString>
                                                <I32>0</I32>
                                            </Obj>
                                            <Obj N="MergePreviousResults" RefId="7">
                                                <TNRef RefId="1"/>
                                                <ToString>None</ToString>
                                                <I32>0</I32>
                                            </Obj>
                                            <Obj N="Args" RefId="8">
                                                <TNRef RefId="0"/>
                                                <LST>
                                                    <Obj RefId="9">
                                                        <MS>
                                                            <S N="N">Name</S>
                                                            <S N="V">var</S>
                                                        </MS>
                                                    </Obj>
                                                    <Obj RefId="10">
                                                        <MS>
                                                            <S N="N">Value</S>
                                                            <S N="V">abc</S>
                                                        </MS>
                                                    </Obj>
                                                </LST>
                                            </Obj>
                                          <Obj N="MergeError" RefId="11">
                                                <TNRef RefId="1"/>
                                                <ToString>None</ToString>
                                                <I32>0</I32>
                                            </Obj>
                                            <Obj N="MergeWarning" RefId="12">
                                                <TNRef RefId="1"/>
                                                <ToString>None</ToString>
                                                <I32>0</I32>
                                            </Obj>
                                            <Obj N="MergeVerbose" RefId="13">
                                                <TNRef RefId="1"/>
                                                <ToString>None</ToString>
                                                <I32>0</I32>
                                            </Obj>
                                            <Obj N="MergeDebug" RefId="14">
                                                <TNRef RefId="1"/>
                                                <ToString>None</ToString>
                                                <I32>0</I32>
                                            </Obj>
                                        </MS>
                                    </Obj>
                                </LST>
                            </Obj>
                        </MS>
                    </Obj>
                    <Obj RefId="15">
                        <MS>
                            <Obj N="Cmds" RefId="16">
                                <TNRef RefId="0"/>
                                <LST>
                                    <Obj RefId="17">
                                        <MS>
                                            <S N="Cmd">Get-Variable</S>
                                            <B N="IsScript">false</B>
                                            <B N="UseLocalScope">false</B>
                                            <Obj N="MergeMyResult" RefId="18">
                                                <TNRef RefId="1"/>
                                                <ToString>None</ToString>
                                                <I32>0</I32>
                                            </Obj>
                                            <Obj N="MergeToResult" RefId="19">
                                                <TNRef RefId="1"/>
                                                <ToString>None</ToString>
                                                <I32>0</I32>
                                            </Obj>
                                            <Obj N="MergePreviousResults" RefId="20">
                                                <TNRef RefId="1"/>
                                                <ToString>None</ToString>
                                                <I32>0</I32>
                                            </Obj>
                                            <Obj N="Args" RefId="21">
                                                <TNRef RefId="0"/>
                                                <LST>
                                                    <Obj RefId="22">
                                                        <MS>
                                                            <S N="N">Name</S>
                                                            <S N="V">var</S>
                                                        </MS>
                                                    </Obj>
                                                </LST>
                                            </Obj>
                                            <Obj N="MergeError" RefId="23">
                                                <TNRef RefId="1"/>
                                                <ToString>None</ToString>
                                                <I32>0</I32>
                                            </Obj>
                                            <Obj N="MergeWarning" RefId="24">
                                                <TNRef RefId="1"/>
                                                <ToString>None</ToString>
                                                <I32>0</I32>
                                            </Obj>
                                            <Obj N="MergeVerbose" RefId="25">
                                                <TNRef RefId="1"/>
                                                <ToString>None</ToString>
                                                <I32>0</I32>
                                            </Obj>
                                            <Obj N="MergeDebug" RefId="26">
                                                <TNRef RefId="1"/>
                                                <ToString>None</ToString>
                                                <I32>0</I32>
                                            </Obj>
                                        </MS>
                                    </Obj>
                                    <Obj RefId="27">
                                        <MS>
                                            <S N="Cmd">Write-Output</S>
                                            <B N="IsScript">false</B>
                                            <B N="UseLocalScope">false</B>
                                            <Obj N="MergeMyResult" RefId="28">
                                                <TNRef RefId="1"/>
                                                <ToString>None</ToString>
                                                <I32>0</I32>
                                            </Obj>
                                            <Obj N="MergeToResult" RefId="29">
                                                <TNRef RefId="1"/>
                                                <ToString>None</ToString>
                                                <I32>0</I32>
                                            </Obj>
                                            <Obj N="MergePreviousResults" RefId="30">
                                                <TNRef RefId="1"/>
                                                <ToString>None</ToString>
                                                <I32>0</I32>
                                            </Obj>
                                            <Obj N="Args" RefId="31">
                                                <TNRef RefId="0"/>
                                                <LST/>
                                            </Obj>
                                            <Obj N="MergeError" RefId="32">
                                                <TNRef RefId="1"/>
                                                <ToString>None</ToString>
                                                <I32>0</I32>
                                            </Obj>
                                            <Obj N="MergeWarning" RefId="33">
                                                <TNRef RefId="1"/>
                                                <ToString>None</ToString>
                                                <I32>0</I32>
                                            </Obj>
                                            <Obj N="MergeVerbose" RefId="34">
                                                <TNRef RefId="1"/>
                                                <ToString>None</ToString>
                                                <I32>0</I32>
                                            </Obj>
                                            <Obj N="MergeDebug" RefId="35">
                                                <TNRef RefId="1"/>
                                                <ToString>None</ToString>
                                                <I32>0</I32>
                                            </Obj>
                                        </MS>
                                    </Obj>
                                </LST>
                            </Obj>
                        </MS>
                    </Obj>
                </LST>
            </Obj>
            <Obj N="Cmds" RefId="36">
                <TNRef RefId="0"/>
                <LST>
                    <Obj RefId="37">
                        <MS>
                            <S N="Cmd">Set-Variable</S>
                            <B N="IsScript">false</B>
                            <B N="UseLocalScope">false</B>
                            <Obj N="MergeMyResult" RefId="38">
                                <TNRef RefId="1"/>
                                <ToString>None</ToString>
                                <I32>0</I32>
                            </Obj>
                            <Obj N="MergeToResult" RefId="39">
                                <TNRef RefId="1"/>
                                <ToString>None</ToString>
                                <I32>0</I32>
                            </Obj>
                            <Obj N="MergePreviousResults" RefId="40">
                                <TNRef RefId="1"/>
                                <ToString>None</ToString>
                                <I32>0</I32>
                            </Obj>
                            <Obj N="Args" RefId="41">
                                <TNRef RefId="0"/>
                                <LST>
                                    <Obj RefId="42">
                                        <MS>
                                            <S N="N">Name</S>
                                            <S N="V">var</S>
                                        </MS>
                                    </Obj>
                                    <Obj RefId="43">
                                        <MS>
                                            <S N="N">Value</S>
                                            <S N="V">abc</S>
                                        </MS>
                                    </Obj>
                                </LST>
                            </Obj>
                            <Obj N="MergeError" RefId="44">
                                <TNRef RefId="1"/>
                                <ToString>None</ToString>
                                <I32>0</I32>
                            </Obj>
                            <Obj N="MergeWarning" RefId="45">
                                <TNRef RefId="1"/>
                                <ToString>None</ToString>
                                <I32>0</I32>
                            </Obj>
                            <Obj N="MergeVerbose" RefId="46">
                                <TNRef RefId="1"/>
                                <ToString>None</ToString>
                                <I32>0</I32>
                            </Obj>
                            <Obj N="MergeDebug" RefId="47">
                                <TNRef RefId="1"/>
                                <ToString>None</ToString>
                                <I32>0</I32>
                            </Obj>
                        </MS>
                    </Obj>
                </LST>
            </Obj>
            <Nil N="History"/>
            <B N="RedirectShellErrorOutputPipe">false</B>
        </MS>
    </Obj>"""

    def test_create_pipeline_single(self):
        serializer = Serializer()

        command = Command(protocol_version="2.2")
        command.cmd = "Set-Variable"
        command.is_script = False
        command.use_local_scope = False
        command.args = [CommandParameter(name="Name", value="var"), CommandParameter(name="Value", value="abc")]
        assert str(command) == "None"

        pipeline = Pipeline()
        pipeline.is_nested = False
        pipeline.commands = [command]
        pipeline.redirect_err_to_out = False

        expected_xml = normalise_xml(self.PIPE_SINGLE)

        actual = serializer.serialize(pipeline)
        actual_xml = normalise_xml(ET.tostring(actual))
        assert_xml_diff(actual_xml, expected_xml)

    def test_create_pipeline_multiple(self):
        serializer = Serializer()

        command1 = Command(protocol_version="2.2")
        command1.cmd = "Set-Variable"
        command1.is_script = False
        command1.use_local_scope = False
        command1.end_of_statement = True
        command1.args = [CommandParameter(name="Name", value="var"), CommandParameter(name="Value", value="abc")]
        assert str(command1) == "None"

        command2 = Command(protocol_version="2.2")
        command2.cmd = "Get-Variable"
        command2.is_script = False
        command2.use_local_scope = False
        command2.args = [
            CommandParameter(name="Name", value="var"),
        ]
        assert str(command2) == "None"

        command3 = Command(protocol_version="2.2")
        command3.cmd = "Write-Output"
        command3.is_script = False
        command3.use_local_scope = False
        assert str(command3) == "None"

        pipeline = Pipeline()
        pipeline.is_nested = False
        pipeline.commands = [command1, command2, command3]
        pipeline.redirect_err_to_out = False

        expected_xml = normalise_xml(self.PIPE_MULTIPLE)

        actual = serializer.serialize(pipeline)
        actual_xml = normalise_xml(ET.tostring(actual))
        assert_xml_diff(actual_xml, expected_xml)

    def test_parse_pipeline_single(self):
        serializer = Serializer()
        actual = serializer.deserialize(normalise_xml(self.PIPE_SINGLE), ObjectMeta("Obj", object=Pipeline))
        assert actual.history is None
        assert actual.is_nested is False
        assert actual.redirect_err_to_out is False
        assert len(actual.commands) == 1
        assert len(actual.commands[0].args) == 2
        assert actual.commands[0].args[0].name == "Name"
        assert actual.commands[0].args[0].value == "var"
        assert actual.commands[0].args[1].name == "Value"
        assert actual.commands[0].args[1].value == "abc"
        assert actual.commands[0].cmd == "Set-Variable"
        assert actual.commands[0].end_of_statement is False
        assert actual.commands[0].is_script is False
        assert str(actual.commands[0].merge_debug) == "None"
        assert str(actual.commands[0].merge_error) == "None"
        assert str(actual.commands[0].merge_my_result) == "None"
        assert str(actual.commands[0].merge_previous) == "None"
        assert str(actual.commands[0].merge_verbose) == "None"
        assert str(actual.commands[0].merge_warning) == "None"
        assert actual.commands[0].use_local_scope is False

    def test_parse_pipeline_multiple(self):
        serializer = Serializer()
        actual = serializer.deserialize(normalise_xml(self.PIPE_MULTIPLE), ObjectMeta("Obj", object=Pipeline))
        assert actual.history is None
        assert actual.is_nested is False
        assert actual.redirect_err_to_out is False
        assert len(actual.commands) == 3
        assert len(actual.commands[0].args) == 2
        assert actual.commands[0].args[0].name == "Name"
        assert actual.commands[0].args[0].value == "var"
        assert actual.commands[0].args[1].name == "Value"
        assert actual.commands[0].args[1].value == "abc"
        assert actual.commands[0].cmd == "Set-Variable"
        assert actual.commands[0].end_of_statement is True
        assert actual.commands[0].is_script is False
        assert str(actual.commands[0].merge_debug) == "None"
        assert str(actual.commands[0].merge_error) == "None"
        assert str(actual.commands[0].merge_my_result) == "None"
        assert str(actual.commands[0].merge_previous) == "None"
        assert str(actual.commands[0].merge_verbose) == "None"
        assert str(actual.commands[0].merge_warning) == "None"
        assert actual.commands[0].use_local_scope is False

        assert len(actual.commands[1].args) == 1
        assert actual.commands[1].args[0].name == "Name"
        assert actual.commands[1].args[0].value == "var"
        assert actual.commands[1].cmd == "Get-Variable"
        assert actual.commands[1].end_of_statement is False
        assert actual.commands[1].is_script is False
        assert str(actual.commands[1].merge_debug) == "None"
        assert str(actual.commands[1].merge_error) == "None"
        assert str(actual.commands[1].merge_my_result) == "None"
        assert str(actual.commands[1].merge_previous) == "None"
        assert str(actual.commands[1].merge_verbose) == "None"
        assert str(actual.commands[1].merge_warning) == "None"
        assert actual.commands[1].use_local_scope is False

        assert len(actual.commands[2].args) == 0
        assert actual.commands[2].cmd == "Write-Output"
        assert actual.commands[2].end_of_statement is True
        assert actual.commands[2].is_script is False
        assert str(actual.commands[2].merge_debug) == "None"
        assert str(actual.commands[2].merge_error) == "None"
        assert str(actual.commands[2].merge_my_result) == "None"
        assert str(actual.commands[2].merge_previous) == "None"
        assert str(actual.commands[2].merge_verbose) == "None"
        assert str(actual.commands[2].merge_warning) == "None"
        assert actual.commands[2].use_local_scope is False


class TestCommandType(object):
    def test_to_string_one_value(self):
        command_type = CommandType(value=1)
        expected = "Alias"
        actual = str(command_type)
        assert actual == expected

    def test_to_string_multiple_values(self):
        command_type = CommandType(value=3)
        expected = "Alias, Function"
        actual = str(command_type)
        assert actual == expected

    def test_to_string_all(self):
        command_type = CommandType(value=CommandType.ALL)
        expected = "All"
        actual = str(command_type)
        assert actual == expected


class TestPipelineResultTypes(object):
    def test_to_string_one_value(self):
        result_type = PipelineResultTypes(value=1)
        expected = "Output"
        actual = str(result_type)
        assert actual == expected

    def test_to_string_multiple_values(self):
        result_type = PipelineResultTypes(value=3)
        expected = "Warning"
        actual = str(result_type)
        assert actual == expected

    def test_to_string_multiple_values_protocol_v2(self):
        result_type = PipelineResultTypes(protocol_version_2=True, value=3)
        expected = "Output, Error"
        actual = str(result_type)
        assert actual == expected

    def test_to_string_none(self):
        result_type = PipelineResultTypes(value=PipelineResultTypes.NONE)
        expected = "None"
        actual = str(result_type)
        assert actual == expected

    def test_to_string_all(self):
        result_type = PipelineResultTypes(value=PipelineResultTypes.ALL)
        expected = "All"
        actual = str(result_type)
        assert actual == expected

    def test_to_string_null(self):
        result_type = PipelineResultTypes(value=PipelineResultTypes.NULL)
        expected = "Null"
        actual = str(result_type)
        assert actual == expected


class TestBufferCell(object):

    BUFFER_CELL = """<Obj RefId="0">
    <Props>
        <C N="character">65</C>
        <Obj N="foregroundColor" RefId="1">
            <TN RefId="0">
                <T>System.ConsoleColor</T>
                <T>System.Enum</T>
                <T>System.ValueType</T>
                <T>System.Object</T>
            </TN>
            <ToString>Cyan</ToString>
            <I32>11</I32>
        </Obj>
        <Obj N="backgroundColor" RefId="2">
            <TNRef RefId="0"/>
            <ToString>Green</ToString>
            <I32>10</I32>
        </Obj>
        <I32 N="bufferCellType">0</I32>
    </Props>
</Obj>"""

    def test_create_buffer_cell(self):
        serializer = Serializer()

        buffer_cell = BufferCell(
            character="A",
            foreground_color=Color(value=Color.CYAN),
            background_color=Color(value=Color.GREEN),
            cell_type=BufferCellType.COMPLETE,
        )

        expected_xml = normalise_xml(self.BUFFER_CELL)
        actual = serializer.serialize(buffer_cell)
        actual_xml = normalise_xml(ET.tostring(actual))

        assert_xml_diff(actual_xml, expected_xml)

    def test_parse_buffer_cell(self):
        serializer = Serializer()
        actual = serializer.deserialize(normalise_xml(self.BUFFER_CELL), ObjectMeta("Obj", object=BufferCell))

        assert actual.character == "A"
        assert actual.foreground_color.value == Color.CYAN
        assert actual.background_color.value == Color.GREEN
        assert actual.cell_type == BufferCellType.COMPLETE


class TestArray(object):

    SINGLE_ARRAY = """<Obj RefId="0">
    <MS>
        <Obj N="mae" RefId="1">
            <TN RefId="0">
                <T>System.Object[]</T>
                <T>System.Array</T>
                <T>System.Object</T>
            </TN>
            <LST>
                <I32>1</I32>
                <I32>2</I32>
                <I32>3</I32>
            </LST>
        </Obj>
        <Obj N="mal" RefId="2">
            <TNRef RefId="0"/>
            <LST>
                <I32>3</I32>
            </LST>
        </Obj>
    </MS>
</Obj>"""

    SINGLE_ARRAY2 = """<Obj RefId="0">
        <MS>
            <Obj N="mae" RefId="1">
                <TN RefId="0">
                    <T>System.Object[]</T>
                    <T>System.Array</T>
                    <T>System.Object</T>
                </TN>
                <LST>
                    <I32>4</I32>
                    <I32>5</I32>
                    <I32>6</I32>
                </LST>
            </Obj>
            <Obj N="mal" RefId="2">
                <TNRef RefId="0"/>
                <LST>
                    <I32>3</I32>
                </LST>
            </Obj>
        </MS>
    </Obj>"""

    TWO_ARRAY = """<Obj RefId="0">
    <MS>
        <Obj N="mae" RefId="1">
            <TN RefId="0">
                <T>System.Object[]</T>
                <T>System.Array</T>
                <T>System.Object</T>
            </TN>
            <LST>
                <I32>1</I32>
                <I32>2</I32>
                <I32>3</I32>
                <I32>4</I32>
                <I32>5</I32>
                <I32>6</I32>
                <I32>7</I32>
                <I32>8</I32>
                <I32>9</I32>
            </LST>
        </Obj>
        <Obj N="mal" RefId="2">
            <TNRef RefId="0"/>
            <LST>
                <I32>3</I32>
                <I32>3</I32>
            </LST>
        </Obj>
    </MS>
</Obj>"""

    THREE_ARRAY = """<Obj RefId="0">
    <MS>
        <Obj N="mae" RefId="1">
            <TN RefId="0">
                <T>System.Object[]</T>
                <T>System.Array</T>
                <T>System.Object</T>
            </TN>
            <LST>
                <I32>1</I32>
                <I32>2</I32>
                <I32>3</I32>
                <I32>4</I32>
                <I32>5</I32>
                <I32>6</I32>
                <I32>7</I32>
                <I32>8</I32>
                <I32>9</I32>
                <I32>10</I32>
                <I32>11</I32>
                <I32>12</I32>
                <I32>13</I32>
                <I32>14</I32>
                <I32>15</I32>
                <I32>16</I32>
                <I32>17</I32>
                <I32>18</I32>
                <I32>19</I32>
                <I32>20</I32>
                <I32>21</I32>
                <I32>22</I32>
                <I32>23</I32>
                <I32>24</I32>
            </LST>
        </Obj>
        <Obj N="mal" RefId="2">
            <TNRef RefId="0"/>
            <LST>
                <I32>2</I32>
                <I32>3</I32>
                <I32>4</I32>
            </LST>
        </Obj>
    </MS>
</Obj>"""

    def test_create_array(self):
        serializer = Serializer()

        array = Array(array=[1, 2, 3])

        expected_xml = normalise_xml(self.SINGLE_ARRAY)
        actual = serializer.serialize(array)
        actual_xml = normalise_xml(ET.tostring(actual))

        assert_xml_diff(actual_xml, expected_xml)

        array.array = [4, 5, 6]
        expected_xml = normalise_xml(self.SINGLE_ARRAY2)
        actual = serializer.serialize(array)
        actual_xml = normalise_xml(ET.tostring(actual))

        assert_xml_diff(actual_xml, expected_xml)

    def test_parse_array(self):
        serializer = Serializer()
        actual = serializer.deserialize(self.SINGLE_ARRAY, ObjectMeta("Obj", object=Array))
        array = actual.array
        assert array == [1, 2, 3]
        assert actual.mae == [1, 2, 3]
        assert actual.mal == [3]

    def test_two_dimensional_create_array(self):
        serializer = Serializer()

        array = Array(array=[[1, 2, 3], [4, 5, 6], [7, 8, 9]])

        expected_xml = normalise_xml(self.TWO_ARRAY)
        actual = serializer.serialize(array)
        actual_xml = normalise_xml(ET.tostring(actual))

        assert_xml_diff(actual_xml, expected_xml)

    def test_parse_two_dimensional_array(self):
        serializer = Serializer()
        actual = serializer.deserialize(self.TWO_ARRAY, ObjectMeta("Obj", object=Array))
        array = actual.array
        assert array == [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        assert actual.mae == [1, 2, 3, 4, 5, 6, 7, 8, 9]
        assert actual.mal == [3, 3]

    def test_three_dimensional_create_array(self):
        serializer = Serializer()

        array = Array(
            array=[
                [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12]],
                [[13, 14, 15, 16], [17, 18, 19, 20], [21, 22, 23, 24]],
            ]
        )

        expected_xml = normalise_xml(self.THREE_ARRAY)
        actual = serializer.serialize(array)
        actual_xml = normalise_xml(ET.tostring(actual))

        assert_xml_diff(actual_xml, expected_xml)

    def test_parse_three_dimensional_array(self):
        serializer = Serializer()
        actual = serializer.deserialize(self.THREE_ARRAY, ObjectMeta("Obj", object=Array))
        array = actual.array
        assert array == [
            [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12]],
            [[13, 14, 15, 16], [17, 18, 19, 20], [21, 22, 23, 24]],
        ]
        assert actual.mae == [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24]
        assert actual.mal == [2, 3, 4]



import collections

import pytest

from pypsrp.encryption import WinRMEncryption
from pypsrp.exceptions import WinRMError

WrapIOVResult = collections.namedtuple("WrapIOVResult", ["buffers"])
WrapResult = collections.namedtuple("WrapResult", ["data"])


class MockAuth:
    def __init__(self, padding=False):
        self.padding = padding

    def wrap_winrm(self, data):
        return b"reallylongheader", data + b"-encrypted", 1 if self.padding else 0

    def unwrap_winrm(self, header, data):
        return data[: len(data) - 10]


class TestWinRMEncryption(object):
    def test_wrap_small_spnego(self):
        plaintext = b"plaintext"
        encryption = WinRMEncryption(MockAuth(), WinRMEncryption.SPNEGO)
        expected = (
            b"--Encrypted Boundary\r\n\tContent-Type: application"
            b"/HTTP-SPNEGO-session-encrypted\r\n\tOriginalContent: "
            b"type=application/soap+xml;charset=UTF-8;Length=9\r\n"
            b"--Encrypted Boundary\r\n\tContent-Type: application/"
            b"octet-stream\r\n\x10\x00\x00\x00reallylongheaderplaintext-"
            b"encrypted--Encrypted Boundary--\r\n"
        )
        actual_type, actual = encryption.wrap_message(plaintext)

        assert "multipart/encrypted" == actual_type
        assert expected == actual

    def test_wrap_spnego_padded(self):
        plaintext = b"plaintext"
        encryption = WinRMEncryption(MockAuth(padding=True), WinRMEncryption.SPNEGO)
        expected = (
            b"--Encrypted Boundary\r\n\tContent-Type: application"
            b"/HTTP-SPNEGO-session-encrypted\r\n\tOriginalContent: "
            b"type=application/soap+xml;charset=UTF-8;Length=10\r\n"
            b"--Encrypted Boundary\r\n\tContent-Type: application/"
            b"octet-stream\r\n\x10\x00\x00\x00reallylongheaderplaintext-"
            b"encrypted--Encrypted Boundary--\r\n"
        )
        actual_type, actual = encryption.wrap_message(plaintext)

        assert "multipart/encrypted" == actual_type
        assert expected == actual

    def test_wrap_small_kerberos(self):
        plaintext = b"plaintext"
        encryption = WinRMEncryption(MockAuth(), WinRMEncryption.KERBEROS)
        expected = (
            b"--Encrypted Boundary\r\n\tContent-Type: application"
            b"/HTTP-Kerberos-session-encrypted\r\n\tOriginalContent: "
            b"type=application/soap+xml;charset=UTF-8;Length=9\r\n"
            b"--Encrypted Boundary\r\n\tContent-Type: application/"
            b"octet-stream\r\n\x10\x00\x00\x00reallylongheaderplaintext-"
            b"encrypted--Encrypted Boundary--\r\n"
        )
        actual_type, actual = encryption.wrap_message(plaintext)

        assert "multipart/encrypted" == actual_type
        assert expected == actual

    def test_wrap_small_credsp(self):
        plaintext = b"plaintext"
        encryption = WinRMEncryption(MockAuth(), WinRMEncryption.CREDSSP)
        expected = (
            b"--Encrypted Boundary\r\n\tContent-Type: application"
            b"/HTTP-CredSSP-session-encrypted\r\n\tOriginalContent: "
            b"type=application/soap+xml;charset=UTF-8;Length=9\r\n"
            b"--Encrypted Boundary\r\n\tContent-Type: application/"
            b"octet-stream\r\n\x10\x00\x00\x00reallylongheaderplaintext-"
            b"encrypted--Encrypted Boundary--\r\n"
        )
        actual_type, actual = encryption.wrap_message(plaintext)

        assert "multipart/encrypted" == actual_type
        assert expected == actual

    def test_wrap_large_spnego(self):
        plaintext = b"a" * 20000
        encryption = WinRMEncryption(MockAuth(), WinRMEncryption.SPNEGO)
        expected = (
            b"--Encrypted Boundary\r\n\tContent-Type: application"
            b"/HTTP-SPNEGO-session-encrypted\r\n\tOriginalContent: "
            b"type=application/soap+xml;charset=UTF-8;Length=20000"
            b"\r\n--Encrypted Boundary\r\n\tContent-Type: application"
            b"/octet-stream\r\n\x10\x00\x00\x00reallylongheader" + plaintext + b"-encrypted--Encrypted Boundary--\r\n"
        )
        actual_type, actual = encryption.wrap_message(plaintext)

        assert "multipart/encrypted" == actual_type
        assert expected == actual

    def test_wrap_large_kerberos(self):
        plaintext = b"a" * 20000
        encryption = WinRMEncryption(MockAuth(), WinRMEncryption.KERBEROS)
        expected = (
            b"--Encrypted Boundary\r\n\tContent-Type: application"
            b"/HTTP-Kerberos-session-encrypted\r\n\tOriginalContent: "
            b"type=application/soap+xml;charset=UTF-8;Length=20000"
            b"\r\n--Encrypted Boundary\r\n\tContent-Type: application"
            b"/octet-stream\r\n\x10\x00\x00\x00reallylongheader" + plaintext + b"-encrypted--Encrypted Boundary--\r\n"
        )
        actual_type, actual = encryption.wrap_message(plaintext)

        assert "multipart/encrypted" == actual_type
        assert expected == actual

    def test_wrap_large_credsp(self):
        plaintext = b"a" * 20000
        encryption = WinRMEncryption(MockAuth(), WinRMEncryption.CREDSSP)
        expected = (
            b"--Encrypted Boundary\r\n\tContent-Type: application"
            b"/HTTP-CredSSP-session-encrypted\r\n\tOriginalContent: "
            b"type=application/soap+xml;charset=UTF-8;Length=16384"
            b"\r\n--Encrypted Boundary\r\n\tContent-Type: application"
            b"/octet-stream\r\n\x10\x00\x00\x00reallylongheader"
            + b"a" * 16384
            + b"-encrypted--Encrypted Boundary\r\n\tContent-Type: "
            b"application/HTTP-CredSSP-session-encrypted\r\n"
            b"\tOriginalContent: type=application/soap+xml;"
            b"charset=UTF-8;Length=3616\r\n--Encrypted Boundary\r\n"
            b"\tContent-Type: application/octet-stream\r\n"
            b"\x10\x00\x00\x00reallylongheader" + b"a" * 3616 + b"-encrypted--Encrypted Boundary--\r\n"
        )
        actual_type, actual = encryption.wrap_message(plaintext)

        assert "multipart/x-multi-encrypted" == actual_type
        assert expected == actual

    def test_unwrap_small_spnego(self):
        expected = b"plaintext"
        encryption = WinRMEncryption(MockAuth(), WinRMEncryption.SPNEGO)
        bwrapped = (
            b"--Encrypted Boundary\r\n\tContent-Type: application"
            b"/HTTP-SPNEGO-session-encrypted\r\n\tOriginalContent: "
            b"type=application/soap+xml;charset=UTF-8;Length=9\r\n"
            b"--Encrypted Boundary\r\n\tContent-Type: application/"
            b"octet-stream\r\n\x10\x00\x00\x00reallylongheaderplaintext-"
            b"encrypted--Encrypted Boundary--\r\n"
        )
        actual = encryption.unwrap_message(bwrapped, "Encrypted Boundary")
        assert expected == actual

    def test_unwrap_small_spnego_without_end_hyphens(self):
        expected = b"plaintext"
        encryption = WinRMEncryption(MockAuth(), WinRMEncryption.SPNEGO)
        bwrapped = (
            b"--Encrypted Boundary\r\n\tContent-Type: application"
            b"/HTTP-SPNEGO-session-encrypted\r\n\tOriginalContent: "
            b"type=application/soap+xml;charset=UTF-8;Length=9\r\n"
            b"--Encrypted Boundary\r\n\tContent-Type: application/"
            b"octet-stream\r\n\x10\x00\x00\x00reallylongheaderplaintext-"
            b"encrypted--Encrypted Boundary\r\n"
        )
        actual = encryption.unwrap_message(bwrapped, "Encrypted Boundary")
        assert expected == actual

    def test_unwrap_small_kerberos(self):
        expected = b"plaintext"
        encryption = WinRMEncryption(MockAuth(), WinRMEncryption.KERBEROS)

        # The spaces after -- on each boundary is on purpose, some MS implementations do this.
        bwrapped = (
            b"-- Encrypted Boundary\r\n\tContent-Type: application"
            b"/HTTP-Kerberos-session-encrypted\r\n\tOriginalContent: "
            b"type=application/soap+xml;charset=UTF-8;Length=9\r\n"
            b"-- Encrypted Boundary\r\n\tContent-Type: application/"
            b"octet-stream\r\n\x10\x00\x00\x00reallylongheaderplaintext-"
            b"encrypted-- Encrypted Boundary--\r\n"
        )
        actual = encryption.unwrap_message(bwrapped, "Encrypted Boundary")
        assert expected == actual

    def test_unwrap_small_credsp(self):
        expected = b"plaintext"
        encryption = WinRMEncryption(MockAuth(), WinRMEncryption.CREDSSP)
        bwrapped = (
            b"--Encrypted Boundary2\r\n\tContent-Type: application"
            b"/HTTP-CredSSP-session-encrypted\r\n\tOriginalContent: "
            b"type=application/soap+xml;charset=UTF-8;Length=9\r\n"
            b"--Encrypted Boundary2\r\n\tContent-Type: application/"
            b"octet-stream\r\n\x10\x00\x00\x00reallylongheaderplaintext-encrypted"
            b"--Encrypted Boundary2--\r\n"
        )
        actual = encryption.unwrap_message(bwrapped, "Encrypted Boundary2")

        assert expected == actual

    def test_unwrap_large_spnego(self):
        expected = b"a" * 20000
        encryption = WinRMEncryption(MockAuth(), WinRMEncryption.SPNEGO)
        bwrapped = (
            b"--Encrypted Boundary\r\n\tContent-Type: application"
            b"/HTTP-SPNEGO-session-encrypted\r\n\tOriginalContent: "
            b"type=application/soap+xml;charset=UTF-8;Length=20000"
            b"\r\n--Encrypted Boundary\r\n\tContent-Type: application"
            b"/octet-stream\r\n\x10\x00\x00\x00reallylongheader" + expected + b"-encrypted--Encrypted Boundary--\r\n"
        )
        actual = encryption.unwrap_message(bwrapped, "Encrypted Boundary")

        assert expected == actual

    def test_unwrap_large_kerberos(self):
        expected = b"a" * 20000
        encryption = WinRMEncryption(MockAuth(), WinRMEncryption.KERBEROS)
        bwrapped = (
            b"--Encrypted Boundary\r\n\tContent-Type: application"
            b"/HTTP-Kerberos-session-encrypted\r\n\tOriginalContent: "
            b"type=application/soap+xml;charset=UTF-8;Length=20000"
            b"\r\n--Encrypted Boundary\r\n\tContent-Type: application"
            b"/octet-stream\r\n\x10\x00\x00\x00reallylongheader" + expected + b"-encrypted--Encrypted Boundary--\r\n"
        )
        actual = encryption.unwrap_message(bwrapped, "Encrypted Boundary")

        assert expected == actual

    def test_unwrap_large_credsp(self):
        expected = b"a" * 20000
        encryption = WinRMEncryption(MockAuth(), WinRMEncryption.CREDSSP)
        bwrapped = (
            b"--Encrypted Boundary\r\n\tContent-Type: application"
            b"/HTTP-CredSSP-session-encrypted\r\n\tOriginalContent: "
            b"type=application/soap+xml;charset=UTF-8;Length=16384"
            b"\r\n--Encrypted Boundary\r\n\tContent-Type: application"
            b"/octet-stream\r\n\x10\x00\x00\x00reallylongheader"
            + b"a" * 16384
            + b"-encrypted--Encrypted Boundary\r\n\tContent-Type: "
            b"application/HTTP-CredSSP-session-encrypted\r\n"
            b"\tOriginalContent: type=application/soap+xml;"
            b"charset=UTF-8;Length=3616\r\n--Encrypted Boundary\r\n"
            b"\tContent-Type: application/octet-stream\r\n"
            b"\x10\x00\x00\x00reallylongheader" + b"a" * 3616 + b"-encrypted--Encrypted Boundary--\r\n"
        )
        actual = encryption.unwrap_message(bwrapped, "Encrypted Boundary")

        assert expected == actual

    def test_unwrap_length_mismatch(self):
        encryption = WinRMEncryption(MockAuth(), WinRMEncryption.SPNEGO)
        bwrapped = (
            b"--Encrypted Boundary\r\n\tContent-Type: application"
            b"/HTTP-SPNEGO-session-encrypted\r\n\tOriginalContent: "
            b"type=application/soap+xml;charset=UTF-8;Length=9\r\n"
            b"--Encrypted Boundary\r\n\tContent-Type: application/"
            b"octet-stream\r\n\x10\x00\x00\x00reallylongheaderplain-"
            b"encrypted--Encrypted Boundary--\r\n"
        )

        with pytest.raises(WinRMError) as err:
            encryption.unwrap_message(bwrapped, "Encrypted Boundary")

        assert (
            str(err.value) == "The encrypted length from the server does not match the "
            "expected length, decryption failed, actual: 5 != expected: 9"
        )



import os

import pytest

from pypsrp.complex_objects import ObjectMeta
from pypsrp.powershell import PowerShell, RunspacePool
from pypsrp.shell import Process, SignalCode, WinRS
from pypsrp.wsman import WSMan


@pytest.fixture(scope="function")
def functional_transports():
    """
    This runs the same test over multiple auth providers and on SSL/Without.
    It requires a specific host configuration of the target endpoint and for
    this config to be defined by the following environment varaibles;

    PYPSRP_RUN_INTEGRATION: if set to anything then these tests will run
    PYPSRP_SERVER: The hostname or IP of the endpoint to test on
    PYPSRP_USERNAME: The username that is an admin on the remote host
    PYPSRP_PASSWORD: THe password for the username
    PYPSRP_CERT_DIR: The directory where the cert.pem and cert_key.pem is
        located for certificate auth. If not defined then certificate auth will
        not be tested

    Here is the test matrix that is run on each test
        http with negotiate
        http with ntlm
        http with credssp
        https with negotiate
        https with ntlm
        https with credssp
        https with basic
        https with cert (depends on env var)

    Each http test has message encryption enabled while the https tests do not.
    """
    run = os.environ.get("PYPSRP_RUN_INTEGRATION", None)
    if run is None:
        pytest.skip("Skipping CI functional tests because PYPSRP_RUN_INTEGRATION has not been set")

    username = os.environ["PYPSRP_USERNAME"]
    password = os.environ["PYPSRP_PASSWORD"]
    server = os.environ["PYPSRP_SERVER"]
    cert_dir = os.environ.get("PYPSRP_CERT_DIR", None)
    http_port = int(os.environ.get("PYPSRP_HTTP_PORT", 5985))
    https_port = int(os.environ.get("PYPSRP_HTTPS_PORT", 5986))

    # can't really test kerberos in CI so it is missing from this list
    auths = ["negotiate", "ntlm", "credssp"]
    auths_ssl = ["basic"]
    if cert_dir is not None:
        auths_ssl.append("certificate")
        cert_key_pem = os.path.join(cert_dir, "cert_key.pem")
        cert_pem = os.path.join(cert_dir, "cert.pem")
    else:
        cert_key_pem = None
        cert_pem = None
    auths_ssl.extend(auths)

    wsmans = []
    for auth in auths:
        wsman = WSMan(server, username=username, password=password, ssl=False, auth=auth, port=http_port)
        wsmans.append(wsman)

    for auth in auths_ssl:
        wsman = WSMan(
            server,
            username=username,
            password=password,
            ssl=True,
            auth=auth,
            cert_validation=False,
            certificate_key_pem=cert_key_pem,
            certificate_pem=cert_pem,
            port=https_port,
        )
        wsmans.append(wsman)
    yield wsmans


class TestPowerShellFunctional(object):
    def test_winrs(self, functional_transports):
        for wsman in functional_transports:
            with wsman, WinRS(wsman) as shell:
                process = Process(shell, "echo", ["hi"])
                process.invoke()
                process.signal(SignalCode.CTRL_C)
                assert process.rc == 0
                assert process.stdout == b"hi\r\n"
                assert process.stderr == b""

    def test_psrp(self, functional_transports):
        for wsman in functional_transports:
            with wsman, RunspacePool(wsman) as pool:
                pool.exchange_keys()
                ps = PowerShell(pool)
                ps.add_cmdlet("Get-Item").add_parameter("Path", "C:\\Windows")
                ps.add_statement()

                sec_string = pool.serialize("super secret", ObjectMeta("SS"))
                ps.add_cmdlet("Set-Variable")
                ps.add_parameter("Name", "password")
                ps.add_parameter("Value", sec_string)

                ps.add_statement().add_script(
                    "[System.Runtime.InteropServices.marshal]"
                    "::PtrToStringAuto([System.Runtime.InteropServices.marshal]"
                    "::SecureStringToBSTR($password))"
                )
                ps.add_statement().add_cmdlet("ConvertTo-SecureString")
                ps.add_parameter("String", "host secret")
                ps.add_parameter("AsPlainText")
                ps.add_parameter("Force")

                large_string = "hello world " * 3000
                ps.add_statement()
                ps.add_script("$VerbosePreference = 'Continue'; Write-Verbose '%s'" % large_string)

                actual = ps.invoke()

            assert ps.had_errors is False
            assert len(actual) == 3
            assert str(actual[0]) == "C:\\Windows"
            assert actual[1] == "super secret"
            assert actual[2] == "host secret"
            assert str(ps.streams.verbose[0]) == large_string

    def test_psrp_jea(self, functional_transports):
        for wsman in functional_transports:
            with wsman, RunspacePool(wsman, configuration_name="JEARole") as pool:
                ps = PowerShell(pool)
                wsman_path = "WSMan:\\localhost\\Service\\AllowUnencrypted"
                ps.add_cmdlet("Get-Item").add_parameter("Path", wsman_path)
                ps.add_statement()
                ps.add_cmdlet("Set-Item").add_parameters({"Path": wsman_path, "Value": "True"})
                actual = ps.invoke()

            assert ps.had_errors is True
            assert len(actual) == 1
            assert actual[0].property_sets[0].adapted_properties["Value"] == "false"
            assert (
                str(ps.streams.error[0]) == "The term 'Set-Item' is not recognized as the name of a "
                "cmdlet, function, script file, or operable program. Check "
                "the spelling of the name, or if a path was included, "
                "verify that the path is correct and try again."
            )
